"""Solvers.

This submodule contains the solvers for the computation of the two-dimensional
thermal processes.

"""

from .explicit_k import explicit_k
from .explicit_general import explicit_general

__all__ = ['explicit_k', 'explicit_general']



"""Heatrapy.

author: Daniel Silva (djsilva@gmx.com)
current version: v2.0.3
url: https://github.com/djsilva99/heatrapy

This module allows to create thermal objects, establish thermal contact between
them, activate or deactivate the whole, or part, of the materials, and compute
the respective heat transfer processes, in 1D and 2D.

The module is based on four main classes:

SingleObject1D
--------------
This class computes the heat transfer processes involved in only one
unidimensional thermal object. It uses the class Object1D changing the state,
modifying the thermal properties and activate or deactivate the material.

SingleObject2D
--------------
This class computes the heat transfer processes involved in only one
two-dimensional thermal object. It uses the class Object2D changing the state,
modifying the thermal properties and activate or deactivate the material.

SystemObjects1D
---------------
This class creates a system of unidimensional thermal objects that can be in
contact to each other and computes the respective heat transfer processes. It
uses the class Object1D changing the state, modifying the thermal properties
and activate or deactivate the material.

SystemObjects2D
---------------
This class creates a system of two-dimensional thermal objects that can be in
contact to each other and computes the respective heat transfer processes. It
uses the class Object2D changing the state, modifying the thermal properties
and activate or deactivate the material.

"""

from .dimension_1.objects import SystemObjects as SystemObjects1D
from .dimension_1.objects import SingleObject as SingleObject1D
from .dimension_2.objects import SystemObjects as SystemObjects2D
from .dimension_2.objects import SingleObject as SingleObject2D

__all__ = ['SystemObjects1D', 'SingleObject1D', 'SystemObjects2D',
           'SingleObject2D']



"""Contains the class single_object.

Used to compute single thermal objects.

"""

from .. import solvers
from . import Object
import matplotlib.pyplot as plt
import numpy as np


class SingleObject:
    """Single_object class.

    This class solves numerically the heat conduction equation for 1 dimension
    of a single material(s). The class has 6 methods.

    """

    def __init__(self, amb_temperature, materials=('Cu',), borders=(1, 11),
                 materials_order=(0,), dx=0.01, dt=0.1, file_name=None,
                 boundaries=(0, 0), initial_state=False,
                 materials_path=False, draw=['temperature'], draw_scale=None):
        """Thermal object initialization.

        `amb_temperature` is the ambient temperature of the whole system.
        `materials` is the list of strings of all the used materials present in
        `material_path`. `borders` is a list of the points where there is a
        change of material. `materials_order` is a list of the materials list
        indexes that defines the material properties given by borders. `dx` and
        `dt` are the space and time steps, respectively. `file_name` is the
        file name where the temperature is saved. `boundaries` is a list of two
        entries that define the boundary condition for temperature. If 0 the
        boundary condition is insulation. `initial_state` is the initial state
        of the materials. True if there are an applied field and False if them
        field is absent. `materials_path` is absolute path of the materials
        database. If false, then the materials database is the standard
        heatrapy database. `draw` is a list of strings representing the online
        plots. In this version only `'temperature'` can be potted. If the list
        is empty, then no drawing is performed. `draw_scale` is a list of two
        values, representing the minimum and maximum temperature to be drawn.
        If None, there are no limits.

        """
        # check the validity of inputs
        materials = tuple(materials)
        borders = tuple(borders)
        materials_order = tuple(materials_order)
        boundaries = tuple(boundaries)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(materials, tuple)
        cond03 = isinstance(borders, tuple)
        cond04 = isinstance(materials_order, tuple)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dt, int) or isinstance(dt, float)
        cond07 = isinstance(file_name, str)
        cond07 = cond07 or (file_name is None)
        cond08 = isinstance(boundaries, tuple)
        cond10 = isinstance(initial_state, bool)
        if isinstance(draw, list):
            cond15 = True
        elif draw is None:
            cond15 = True
        else:
            cond15 = False
        if isinstance(draw_scale, list) or isinstance(draw_scale, tuple):
            cond16 = (len(draw_scale) == 2)
        elif draw_scale is None:
            cond16 = True
        else:
            cond16 = False
        condition = cond01 and cond02 and cond03 and cond04 and cond05
        condition = condition and cond06 and cond07 and cond08
        condition = condition and cond10
        condition = condition and cond15 and cond16
        if not condition:
            raise ValueError

        self.object = Object(amb_temperature, materials=materials,
                             borders=borders, materials_order=materials_order,
                             dx=dx, dt=dt, file_name=file_name,
                             boundaries=boundaries,
                             initial_state=initial_state,
                             materials_path=materials_path)

        # initializes the plotting
        self.draw = draw
        self.draw_scale = draw_scale
        for drawing in self.draw:
            if drawing == 'temperature':
                self.figure = plt.figure()
                self.ax = self.figure.add_subplot(111)
                temp = []
                for i in range(len(self.object.temperature)):
                    temp.append(self.object.temperature[i][0])
                if not self.draw_scale:
                    vmax = max(temp)
                    vmin = min(temp)
                    if vmax == vmin:
                        vmin = vmin - 0.1
                        vmax = vmax + 0.1
                    temp = np.array(temp)
                    x_plot = [self.object.dx*j for j in range(len(temp))]
                    self.online, = self.ax.plot(x_plot, temp)
                    self.ax.set_ylim([vmin, vmax])
                else:
                    temp = np.array(temp)
                    x_plot = [self.object.dx*j for j in range(len(temp))]
                    self.online, = self.ax.plot(x_plot, temp)
                    self.ax.set_ylim(self.draw_scale)
                self.ax.set_title('Temperature (K)')
                self.ax.set_xlabel('x axis (m)')
                self.ax.set_ylabel('temperature (K)')
                plt.show(block=False)

    def show_figure(self, figure_type, draw_scale=None):
        """Plotting.

        Initializes a specific live plotting. `figure_type` is a string
        identifying the plotting. This version only allows the plotting of the
        'temperature'. `draw_scale` defines the range of temperatures. If None,
        this range is found automatically for every frame.

        """
        # check the validity of inputs
        if isinstance(draw_scale, list) or isinstance(draw_scale, tuple):
            condition = (len(draw_scale) == 2)
        elif draw_scale is None:
            condition = True
        else:
            condition = False
        condition = condition and isinstance(figure_type, str)
        if not condition:
            raise ValueError

        self.draw_scale = draw_scale
        if figure_type == 'temperature':
            if figure_type not in self.draw:
                self.draw.append(figure_type)
            self.figure = plt.figure()
            self.ax = self.figure.add_subplot(111)
            temp = []
            for i in range(len(self.object.temperature)):
                temp.append(self.object.temperature[i][0])
            if not self.draw_scale:
                vmax = max(temp)
                vmin = min(temp)
                if vmax == vmin:
                    vmin = vmin - 0.1
                    vmax = vmax + 0.1
                temp = np.array(temp)
                x_plot = [self.object.dx*j for j in range(len(temp))]
                self.online, = self.ax.plot(x_plot, temp)
                self.ax.set_ylim([vmin, vmax])
            else:
                temp = np.array(temp)
                x_plot = [self.object.dx*j for j in range(len(temp))]
                self.online, = self.ax.plot(x_plot, temp)
                self.ax.set_ylim(self.draw_scale)
            self.ax.set_title('Temperature (K)')
            self.ax.set_xlabel('x axis (m)')
            self.ax.set_ylabel('temperature (K)')
            plt.show(block=False)

    def activate(self, initial_point, final_point):
        """Activation.

        Activates the thermal object between `initial_point` to `final_point`.
        """
        # check the validity of inputs
        condition = isinstance(initial_point, int)
        condition = condition and isinstance(final_point, int)
        if not condition:
            raise ValueError

        self.object.activate(initial_point, final_point)

        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    try:
                        temp = []
                        for i in range(len(self.object.temperature)):
                            temp.append(self.object.temperature[i][0])
                        if not self.draw_scale:
                            vmax = max(temp)
                            vmin = min(temp)
                            if vmax == vmin:
                                vmin = vmin - 0.1
                                vmax = vmax + 0.1
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                            self.ax.set_ylim([vmin, vmax])
                        else:
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                        self.figure.canvas.draw()
                    except Exception:
                        pass

    def deactivate(self, initial_point, final_point):
        """Deactivation.

        Deactivates the thermal object between `initial_point` to
        `final_point`.
        """
        # check the validity of inputs
        condition = isinstance(initial_point, int)
        condition = condition and isinstance(final_point, int)
        if not condition:
            raise ValueError

        self.object.deactivate(initial_point, final_point)

        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    try:
                        temp = []
                        for i in range(len(self.object.temperature)):
                            temp.append(self.object.temperature[i][0])
                        if not self.draw_scale:
                            vmax = max(temp)
                            vmin = min(temp)
                            if vmax == vmin:
                                vmin = vmin - 0.1
                                vmax = vmax + 0.1
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                            self.ax.set_ylim([vmin, vmax])
                        else:
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                        self.figure.canvas.draw()
                    except Exception:
                        pass

    def change_power(self, power_type, power, initial_point, final_point):
        """Heat power source change.

        Changes the coeficients for the heat power sources by a value of power
        from `initial_point` to `final_point`. `power_type` is a string that
        represents the type of coefficient, i.e. 'Q' or 'Q0'.

        """
        # check the validity of inputs
        value = isinstance(initial_point, int)
        if value and isinstance(final_point, int):
            cond1 = True
        else:
            cond1 = False
        cond2 = isinstance(power, int) or isinstance(power, float)
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        if not (cond1 and cond2 and cond3):
            raise ValueError

        if power_type == 'Q':
            for j in range(initial_point, final_point):
                self.object.Q[j] = power
        if power_type == 'Q0':
            for j in range(initial_point, final_point):
                self.object.Q0[j] = power

    def change_boundaries(self, boundaries):
        """Boundary change.

        Changes the `boundaries` variable.

        """
        # check the validity of inputs
        if isinstance(boundaries, tuple):
            if len(boundaries) == 2:
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.boundaries = boundaries

    def compute(self, time_interval, write_interval, solver='explicit_k(x)',
                verbose=True):
        """Compute the thermal process.

        Computes the system for time_interval seconds, and writes into the
        `file_name` file every `write_interval` time steps. Four different
        solvers can be used: `'explicit_general'`, `'explicit_k(x)'`,
        `'implicit_general'`, and `'implicit_k(x)'`. If `verbose = True`, then
        the progress of the computation progress is shown.

        """
        # check the validity of inputs
        cond1 = isinstance(time_interval, float)
        cond1 = cond1 or isinstance(time_interval, int)
        cond2 = isinstance(write_interval, int)
        if isinstance(solver, str):
            all_solvers = ['implicit_general', 'implicit_k(x)',
                           'explicit_k(x)', 'explicit_general']
            if solver in all_solvers:
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        cond4 = isinstance(verbose, bool)
        condition = cond1 and cond2 and cond3 and cond4
        if not condition:
            raise ValueError

        # number of time steps for the given timeInterval
        nt = int(time_interval / self.object.dt)

        # number of time steps counting from the last writing process
        nw = 0

        # computes
        for j in range(nt):

            # updates the time_passed
            self.object.time_passed = self.object.time_passed + self.object.dt

            # defines the material properties accoring to the state list
            for i in range(1, self.object.num_points - 1):
                if self.object.state[i] is True:
                    value = self.object.materials_index[i]
                    self.object.rho[i] = self.object.materials[value].rhoa(
                        self.object.temperature[i][0])
                    self.object.Cp[i] = self.object.materials[value].cpa(
                        self.object.temperature[i][0])
                    self.object.k[i] = self.object.materials[value].ka(
                        self.object.temperature[i][0])
                if self.object.state[i] is False:
                    value = self.object.materials_index[i]
                    self.object.rho[i] = self.object.materials[value].rho0(
                        self.object.temperature[i][0])
                    self.object.Cp[i] = self.object.materials[value].cp0(
                        self.object.temperature[i][0])
                    self.object.k[i] = self.object.materials[value].k0(
                        self.object.temperature[i][0])

            # SOLVERS
            # implicit k constant
            if solver == 'implicit_general':
                value = solvers.implicit_general(self.object)
                self.object.temperature, self.object.lheat = value
            # implicit k dependent on x
            if solver == 'implicit_k(x)':
                value = solvers.implicit_k(self.object)
                self.object.temperature, self.object.lheat = value
            # explicit k constant
            if solver == 'explicit_general':
                value = solvers.explicit_general(self.object)
                self.object.temperature, self.object.lheat = value
            # explicit k dependent on x
            if solver == 'explicit_k(x)':
                value = solvers.explicit_k(self.object)
                self.object.temperature, self.object.lheat = value

            nw = nw + 1

            if self.draw:
                for drawing in self.draw:
                    if drawing == 'temperature':
                        try:
                            value = nw + 1 == write_interval
                            if value or j == 0 or j == nt - 1:
                                temp = []
                                for i in range(len(self.object.temperature)):
                                    temp.append(self.object.temperature[i][0])
                                if not self.draw_scale:
                                    vmax = max(temp)
                                    vmin = min(temp)
                                    if vmax == vmin:
                                        vmin = vmin - 0.1
                                        vmax = vmax + 0.1
                                    temp = np.array(temp)
                                    self.online.set_ydata(temp)
                                    self.ax.set_ylim([vmin, vmax])
                                else:
                                    temp = np.array(temp)
                                    self.online.set_ydata(temp)
                                self.figure.canvas.draw()
                        except Exception:
                            pass

            # writes the temperature to file_name file ...
            # if the number of time steps is verified
            if self.object.file_name:
                if nw == write_interval or j == 0 or j == nt - 1:
                    line = '%f,' % self.object.time_passed
                    for i in self.object.temperature:
                        new_line = '%f,' % i[1]
                        line = line + new_line
                    line = line[:-1] + '\n'
                    f = open(self.object.file_name, 'a')
                    f.write(line)
                    f.close()

            if nw == write_interval:
                nw = 0
                if verbose:
                    print('progress:', int(100*j/nt), '%', end="\r")

        if verbose:
            print('Finished simulation')



"""Contains the class single_object.

Used to compute single two-dimensional system models.

"""

import copy
from .. import solvers
from . import Object
import numpy as np
import matplotlib.pyplot as plt
import matplotlib


class SingleObject:
    """Single_object class.

    This class solves numerically the two-dimensional heat conduction equation.

    """

    def __init__(self, amb_temperature, material='Cu', dx=0.01, dy=0.01,
                 dt=0.1, size=(10, 10), file_name=None,
                 boundaries=(0, 0, 0, 0), Q=[], Q0=[], initial_state=False,
                 materials_path=False,
                 draw=['temperature', 'materials'], draw_scale=None):
        """Object initialization.

        `amb_temperature` is the ambient temperature of the whole system.
        `materials` is the background material present in `material_path`.
        `dx`, `dy` are the space steps along the x- and y-axis, respectively.
        `dt` is the time step. `file_name` is the file name where the
        temperature is saved. `boundaries` is a list of four entries that
        define the boundary condition for temperature (left, right, bottom,
        top). If 0 the boundary condition is insulation. `initial_state` is the
        initial state of the materials. True if there are an applied field and
        False if them field is absent. `materials_path` is absolute path of the
        materials database. If false, then the materials database is the
        standard heatrapy database. `draw` is a list of strings representing
        the online plots. In this version live plotting can be performed for
        `temperature`, `materials`, `state`, `Q` and `Q0`. If the list is
        empty, then no drawing is performed. `draw_scale` is a list of two
        values, representing the minimum and maximum temperature to be drawn.
        If None, there are no limits. `Q` is a list of fixed heat source
        coefficient and `Q0` is a list of temperature dependent heat source
        coefficient.

        """
        boundaries = tuple(boundaries)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(material, str)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dy, int) or isinstance(dy, float)
        cond07 = isinstance(dt, int) or isinstance(dt, float)
        cond08 = isinstance(file_name, str)
        cond08 = cond08 or (file_name is None)
        cond09 = isinstance(boundaries, tuple)
        cond10 = isinstance(initial_state, bool)
        cond11 = isinstance(Q, list)
        cond12 = isinstance(Q0, list)
        cond13 = isinstance(draw, list)
        cond14 = isinstance(draw_scale, list) or isinstance(draw_scale, tuple)
        cond14 = cond14 or draw_scale is None
        condition = cond01 and cond02 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        condition = condition and cond10 and cond11 and cond12 and cond13
        condition = condition and cond14
        if not condition:
            raise ValueError

        self.materials_path = materials_path
        self.time_passed = 0.
        self.size = size
        self.dt = dt
        self.dx = dx
        self.dy = dy
        self.object = Object(amb_temperature, material=material, dx=dx, dy=dy,
                             dt=dt, size=size, file_name=file_name,
                             boundaries=boundaries, Q=[], Q0=[],
                             initial_state=initial_state,
                             materials_path=materials_path)

        # Initializes plotting
        self.draw = draw
        self.draw_scale = draw_scale
        if self.draw:
            cmap = matplotlib.cm.RdBu
            name = 'my_cmap_r'
            reverse = []
            k = []
            for key in cmap._segmentdata:
                k.append(key)
                channel = cmap._segmentdata[key]
                data = []
                for t in channel:
                    data.append((1-t[0], t[2], t[1]))
                reverse.append(sorted(data))
            linear = dict(zip(k, reverse))
            my_cmap_r = matplotlib.colors.LinearSegmentedColormap(name, linear)
            cmap_r = my_cmap_r
            for drawing in self.draw:
                if drawing == 'temperature':
                    self.figure = plt.figure()
                    self.ax = self.figure.add_subplot(111)
                    temp = []
                    for i in range(self.object.size[0]):
                        temp.append([])
                        for j in range(self.object.size[1]):
                            temp[-1].append(self.object.temperature[i][j][0])
                    if not self.draw_scale:
                        vmax = max(max(temp, key=max))
                        vmin = min(min(temp, key=min))
                        temp = np.array(temp)
                        extent = [0, size[0]*dx, 0, size[1]*dy]
                        self.im = self.ax.imshow(np.transpose(temp), vmax=vmax,
                                                 vmin=vmin, cmap=cmap_r,
                                                 extent=extent, origin='lower',
                                                 interpolation='hamming')
                    else:
                        temp = np.array(temp)
                        extent = [0, size[0]*dx, 0, size[1]*dy]
                        self.im = self.ax.imshow(np.transpose(temp),
                                                 vmax=self.draw_scale[0],
                                                 vmin=self.draw_scale[1],
                                                 cmap=cmap_r, extent=extent,
                                                 origin='lower',
                                                 interpolation='hamming')
                    cbar_kw = {}
                    self.ax.figure.colorbar(
                        self.im,
                        ax=self.ax,
                        **cbar_kw
                    )
                    self.ax.set_title('Temperature (K)')
                    self.ax.set_xlabel('x axis (m)')
                    self.ax.set_ylabel('y axis (m)')
                    plt.show(block=False)
                if drawing == 'state':
                    self.figure_state = plt.figure()
                    self.ax_state = self.figure_state.add_subplot(111)
                    vmax = 1
                    vmin = 0
                    state = np.array(self.object.state)
                    cmap = plt.get_cmap("gray", 2)
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    self.im_state = self.ax_state.imshow(np.transpose(state),
                                                         vmax=1.5, vmin=-0.5,
                                                         cmap=cmap,
                                                         extent=extent,
                                                         origin='lower')
                    cbar_kw = {}
                    cbarlabel = ""
                    qrates = np.array(['active', 'inactive'])
                    value = np.linspace(0, 1, 2)
                    norm = matplotlib.colors.BoundaryNorm(value, 1)
                    func = lambda x, pos: qrates[::-1][norm(x)]  # noqa: E731
                    fmt = matplotlib.ticker.FuncFormatter(func)
                    cbar_kw = dict(ticks=np.arange(-3, 4), format=fmt)
                    cbar_stat = self.ax_state.figure.colorbar(self.im_state,
                                                              ax=self.ax_state,
                                                              **cbar_kw)
                    cbar_stat.ax.set_ylabel(cbarlabel, rotation=-90,
                                            va="bottom")
                    self.ax_state.set_title('State')
                    self.ax_state.set_xlabel('x axis (m)')
                    self.ax_state.set_ylabel('y axis (m)')
                    plt.show(block=False)
                value = len(self.object.materials_name) > 1
                if drawing == 'materials' and value:
                    self.figure_materials = plt.figure()
                    self.ax_materials = self.figure_materials.add_subplot(111)
                    vmax = len(self.object.materials)-1
                    vmin = 0
                    material_id = np.array(self.object.materials_index)
                    cmap = plt.get_cmap("gray", vmax)
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    origin = 'lower'
                    value = np.transpose(material_id)
                    self.im_materials = self.ax_materials.imshow(value,
                                                                 vmax=vmax,
                                                                 vmin=vmin,
                                                                 cmap=cmap,
                                                                 extent=extent,
                                                                 origin=origin)
                    cbar_kw = {}
                    cbarlabel = ""
                    qrates = np.array(self.object.materials_name)
                    value_1 = np.linspace(0, 1, len(self.object.materials))
                    value_2 = len(self.object.materials)-1
                    norm = matplotlib.colors.BoundaryNorm(value_1, value_2)
                    func = lambda x, pos: qrates[::-1][norm(x)]  # noqa: E731
                    fmt = matplotlib.ticker.FuncFormatter(func)
                    cbar_kw = dict(ticks=np.arange(-3, 4), format=fmt)
                    value_1 = self.im_materials
                    value_2 = self.ax_materials
                    cbar_m = self.ax_materials.figure.colorbar(value_1,
                                                               ax=value_2,
                                                               **cbar_kw)
                    cbar_m.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
                    self.ax_materials.set_title('Materials')
                    self.ax_materials.set_xlabel('x axis (m)')
                    self.ax_materials.set_ylabel('y axis (m)')
                    plt.show(block=False)
                if drawing == 'Q':
                    self.figure_Q = plt.figure()
                    self.ax_Q = self.figure_Q.add_subplot(111)
                    temp = np.array(self.object.Q)
                    vmax = 1
                    vmin = 0
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    self.im_Q = self.ax_Q.imshow(np.transpose(temp), vmax=vmax,
                                                 vmin=vmin, cmap='inferno',
                                                 extent=extent, origin='lower',
                                                 interpolation='hamming')
                    cbar_kw = {}
                    self.ax_Q.figure.colorbar(
                        self.im_Q,
                        ax=self.ax_Q,
                        **cbar_kw
                    )
                    self.ax_Q.set_title('Q (W/m²)')
                    self.ax_Q.set_xlabel('x axis (m)')
                    self.ax_Q.set_ylabel('y axis (m)')
                    plt.show(block=False)
                if drawing == 'Q0':
                    self.figure_Q0 = plt.figure()
                    self.ax_Q0 = self.figure_Q0.add_subplot(111)
                    temp = np.array(self.object.Q0)
                    vmax = 1
                    vmin = 0
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    self.im_Q0 = self.ax_Q0.imshow(np.transpose(temp),
                                                   vmax=vmax, vmin=vmin,
                                                   cmap='inferno',
                                                   extent=extent,
                                                   origin='lower',
                                                   interpolation='hamming')
                    cbar_kw = {}
                    self.ax_Q0.figure.colorbar(
                        self.im_Q0,
                        ax=self.ax_Q0,
                        **cbar_kw
                    )
                    self.ax_Q0.set_title('Q0 (W/m²)')
                    self.ax_Q0.set_xlabel('x axis (m)')
                    self.ax_Q0.set_ylabel('y axis (m)')
                    plt.show(block=False)

    def show_figure(self, figure_type):
        """Plotting.

        Initializes a specific plotting. `figure_type` is a string identifying
        the plotting. In this version live plotting can be performed for
        `temperature`, `materials`, `state`, `Q` and `Q0`.

        """
        # check the validity of inputs
        condition = isinstance(figure_type, str)
        if not condition:
            raise ValueError

        if figure_type not in self.draw:
            self.draw.append(figure_type)
        if figure_type == 'temperature':
            self.figure = plt.figure()
            self.ax = self.figure.add_subplot(111)
            temp = []
            for i in range(self.object.size[0]):
                temp.append([])
                for j in range(self.object.size[1]):
                    temp[-1].append(self.object.temperature[i][j][0])
            if not self.draw_scale:
                vmax = max(max(temp, key=max))
                vmin = min(min(temp, key=min))
                temp = np.array(temp)
                extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
                self.im = self.ax.imshow(np.transpose(temp), vmax=vmax,
                                         vmin=vmin, cmap='jet', extent=extent,
                                         origin='lower',
                                         interpolation='hamming')
            else:
                temp = np.array(temp)
                extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
                self.im = self.ax.imshow(np.transpose(temp),
                                         vmax=self.draw_scale[0],
                                         vmin=self.draw_scale[1],
                                         cmap='jet', extent=extent,
                                         origin='lower',
                                         interpolation='hamming')
            cbar_kw = {}
            cbarlabel = "temperature (K)"
            cbar = self.ax.figure.colorbar(self.im, ax=self.ax, **cbar_kw)
            cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
            self.ax.set_title('Temperature')
            self.ax.set_xlabel('x axis (m)')
            self.ax.set_ylabel('y axis (m)')
            plt.show(block=False)
        if figure_type == 'state':
            self.figure_state = plt.figure()
            self.ax_state = self.figure_state.add_subplot(111)
            vmax = 1
            vmin = 0
            state = np.array(self.object.state)
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            self.im_state = self.ax_state.imshow(np.transpose(state), vmax=1.5,
                                                 vmin=-0.5,
                                                 cmap=plt.get_cmap("gray", 2),
                                                 extent=extent, origin='lower')
            cbar_kw = {}
            cbarlabel = ""  # "state"
            qrates = np.array(['active', 'inactive'])
            norm = matplotlib.colors.BoundaryNorm(np.linspace(0, 1, 2), 1)
            func = lambda x, pos: qrates[::-1][norm(x)]  # noqa: E731
            fmt = matplotlib.ticker.FuncFormatter(func)
            cbar_kw = dict(ticks=np.arange(-3, 4), format=fmt)
            cbar_state = self.ax_state.figure.colorbar(self.im_state,
                                                       ax=self.ax_state,
                                                       **cbar_kw)
            cbar_state.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
            self.ax_state.set_title('State')
            self.ax_state.set_xlabel('x axis (m)')
            self.ax_state.set_ylabel('y axis (m)')
            plt.show(block=False)
        if figure_type == 'materials' and len(self.object.materials_name) > 1:
            self.figure_materials = plt.figure()
            self.ax_materials = self.figure_materials.add_subplot(111)
            vmax = len(self.object.materials)-1
            vmin = 0
            material_id = np.array(self.object.materials_index)
            cmap = plt.get_cmap("PiYG", vmax+1)
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            value = np.transpose(material_id)
            self.im_materials = self.ax_materials.imshow(value,
                                                         vmax=vmax,
                                                         vmin=vmin,
                                                         cmap=cmap,
                                                         extent=extent,
                                                         origin='lower')
            cbar_kw = {}
            cbarlabel = ""
            materials_name_list = copy.deepcopy(self.object.materials_name)
            materials_name_list.reverse()
            qrates = np.array(materials_name_list)
            value = np.linspace(0, len(self.object.materials) - 1,
                                len(self.object.materials))
            norm = matplotlib.colors.BoundaryNorm(value,
                                                  len(self.object.materials)-1)
            func = lambda x, pos: qrates[::-1][norm(x)]  # noqa: E731
            fmt = matplotlib.ticker.FuncFormatter(func)
            cbar_kw = dict(ticks=np.arange(0, len(self.object.materials)+1),
                           format=fmt)
            cbar_m = self.ax_materials.figure.colorbar(self.im_materials,
                                                       ax=self.ax_materials,
                                                       **cbar_kw)
            cbar_m.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
            self.ax_materials.set_title('Materials')
            self.ax_materials.set_xlabel('x axis (m)')
            self.ax_materials.set_ylabel('y axis (m)')
            plt.show(block=False)
            print(self.object.materials_name)
        if figure_type == 'materials' and len(self.object.materials_name) == 1:
            print(self.object.materials_name)
        if figure_type == 'Q':
            self.figure_Q = plt.figure()
            self.ax_Q = self.figure_Q.add_subplot(111)
            temp = np.array(self.object.Q)
            vmax = 1
            vmin = 0
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            self.im_Q = self.ax_Q.imshow(np.transpose(temp), vmax=vmax,
                                         vmin=vmin, cmap='inferno',
                                         extent=extent, origin='lower',
                                         interpolation='hamming')
            cbar_kw = {}
            cbar = self.ax_Q.figure.colorbar(self.im_Q, ax=self.ax_Q,
                                             **cbar_kw)
            self.ax_Q.set_title('Q (W/m²)')
            self.ax_Q.set_xlabel('x axis (m)')
            self.ax_Q.set_ylabel('y axis (m)')
            plt.show(block=False)
        if figure_type == 'Q0':
            self.figure_Q0 = plt.figure()
            self.ax_Q0 = self.figure_Q0.add_subplot(111)
            temp = np.array(self.object.Q0)
            vmax = 1
            vmin = 0
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            self.im_Q0 = self.ax_Q0.imshow(np.transpose(temp), vmax=vmax,
                                           vmin=vmin, cmap='inferno',
                                           extent=extent,
                                           origin='lower',
                                           interpolation='hamming')
            cbar_kw = {}
            cbar = self.ax_Q0.figure.colorbar(self.im_Q0, ax=self.ax_Q0,
                                              **cbar_kw)
            self.ax_Q0.set_title('Q0 (W/m²)')
            self.ax_Q0.set_xlabel('x axis (m)')
            self.ax_Q0.set_ylabel('y axis (m)')
            plt.show(block=False)

    def activate(self, initial_point, final_point, shape='square'):
        """Activation of the material.

        Activates a given piece of material. If `shape` is `'square'`, then the
        `initial_point` is the tuple (x,y) of the bottom left point and the
        `final_point` is the tuple (x,y) of the top right point. If the shape
        is `'circle'`, the `initial_point` is the tuple (x,y) of the center of
        the circle and `final_point` is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.activate(initial_point=initial_point,
                             final_point=final_point, shape=shape)
        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    temp = []
                    for i in range(self.object.size[0]):
                        temp.append([])
                        for j in range(self.object.size[1]):
                            temp[-1].append(self.object.temperature[i][j][0])
                    self.im.set_array(np.transpose(temp))
                    if not self.draw_scale:
                        vmax = max(max(temp, key=max))
                        vmin = min(min(temp, key=min))
                        self.im.set_clim(vmin=vmin)
                        self.im.set_clim(vmax=vmax)
                    self.figure.canvas.draw()
                if drawing == 'state':
                    self.im_state.set_array(np.transpose(self.object.state))
                    self.figure_state.canvas.draw()

    def deactivate(self, initial_point, final_point, shape='square'):
        """Deactivation of the material.

        Deactivates a given piece of material. If `shape` is `'square'`, then
        the `initial_point` is the tuple (x,y) of the bottom left point and the
        `final_point` is the tuple (x,y) of the top right point. If the shape
        is `'circle'`, the `initial_point` is the tuple (x,y) of the center of
        the circle and `final_point` is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.deactivate(initial_point=initial_point,
                               final_point=final_point, shape=shape)
        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    temp = []
                    for i in range(self.object.size[0]):
                        temp.append([])
                        for j in range(self.object.size[1]):
                            temp[-1].append(self.object.temperature[i][j][0])
                    self.im.set_array(np.transpose(temp))
                    if not self.draw_scale:
                        vmax = max(max(temp, key=max))
                        vmin = min(min(temp, key=min))
                        self.im.set_clim(vmin=vmin)
                        self.im.set_clim(vmax=vmax)
                    self.figure.canvas.draw()
                if drawing == 'state':
                    self.im_state.set_array(np.transpose(self.object.state))
                    self.figure_state.canvas.draw()

    def change_boundaries(self, boundaries):
        """Boundary change.

        Changes the `boundaries` variable.

        """
        # check the validity of inputs
        if isinstance(boundaries, tuple):
            if len(boundaries) == 4:
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.boundaries = boundaries

    def change_material(self, shape, material, initial_point, length,
                        state=False):
        """Material change.

        Changes the material of a given piece of the background material. If
        `shape` is `'square'`, then the initial_point is the tuple (x,y) of the
        bottom left point and the `length` is the tuple (x,y) of the length. If
        the shape is `'circle'`, the `initial_point` is the tuple (x,y) of the
        center of the circle and `length` is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(length, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(length) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(length, int)
                value = value or isinstance(length, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        condition = condition and isinstance(state, bool)
        if not condition:
            raise ValueError

        if shape == 'square':
            self.object.square(material=material,
                               initial_point=initial_point, length=length,
                               state=state, materials_path=self.materials_path)
        if shape == 'circle':
            self.object.circle(material=material,
                               initial_point=initial_point, radius=length,
                               state=state, materials_path=self.materials_path)
        try:
            plt.close(self.figure_materials)
        except Exception:
            pass

        if self.draw:
            if 'materials' in self.draw:
                self.figure_materials = plt.figure()
                self.ax_materials = self.figure_materials.add_subplot(111)
                vmax = len(self.object.materials)-1
                vmin = 0
                cmap = plt.get_cmap("PiYG", vmax+1)
                extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
                material_id = np.array(self.object.materials_index)
                value = np.transpose(material_id)
                self.im_materials = self.ax_materials.imshow(value,
                                                             vmax=vmax,
                                                             vmin=vmin,
                                                             cmap=cmap,
                                                             extent=extent,
                                                             origin='lower')
                cbar_kw = {}
                cbarlabel = ""
                materials_name_list = copy.deepcopy(self.object.materials_name)
                materials_name_list.reverse()
                qrates = np.array(materials_name_list)
                value_1 = np.linspace(0, len(self.object.materials)-1,
                                      len(self.object.materials))
                value_2 = len(self.object.materials)-1
                norm = matplotlib.colors.BoundaryNorm(value_1, value_2)
                func = lambda x, pos: qrates[::-1][norm(x)]  # noqa: E731
                fmt = matplotlib.ticker.FuncFormatter(func)
                value = np.arange(0, len(self.object.materials)+1)
                cbar_kw = dict(ticks=value, format=fmt)
                cba_m = self.ax_materials.figure.colorbar(self.im_materials,
                                                          ax=self.ax_materials,
                                                          **cbar_kw)
                cba_m.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
                self.ax_materials.set_title('Materials')
                self.ax_materials.set_xlabel('x axis (m)')
                self.ax_materials.set_ylabel('y axis (m)')
                plt.show(block=False)

    def change_power(self, shape, power_type, initial_point, final_point,
                     power):
        """Power change.

        Changes the power matrix of the thermal object. If `shape` is
        `'square'`, then the `initial_point` is the tuple (x,y) of the bottom
        left point and the `final_point` is the tuple (x,y) of the top right
        point. If the `shape` is `'circle'`, the `initial_point` is the tuple
        (x,y) of the center of the circle and `final_point` is its radius.
        `power` is the value of the power to add, and `power_type` is the type
        of power to be introduced, which has the value `'Q'` if it is
        temperature dependent and `'Q0'` if it is temperature independent.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    cond1 = len(initial_point) == 2
                    cond1 = cond1 and len(final_point) == 2
                else:
                    cond1 = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    cond1 = len(initial_point) == 2
                else:
                    cond1 = False
            else:
                cond1 = False
        else:
            cond1 = False
        cond2 = isinstance(power, int) or isinstance(power, float)
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        if not cond1 and cond2 and cond3:
            raise ValueError

        if shape == 'square':
            self.object.power_add(initial_point, final_point, power,
                                  shape=shape, power_type=power_type)
        if shape == 'circle':
            self.object.power_add(initial_point, final_point, power,
                                  shape=shape, power_type=power_type)
        if self.draw:
            if 'Q' in self.draw and power_type == 'Q':
                self.im_Q.set_array(np.transpose(self.object.Q))
                vmax = max(max(self.object.Q, key=max))
                vmin = min(min(self.object.Q, key=min))
                self.im_Q.set_clim(vmin=vmin)
                self.im_Q.set_clim(vmax=vmax)
                self.figure_Q.canvas.draw()
            if 'Q0' in self.draw and power_type == 'Q0':
                self.im_Q0.set_array(np.transpose(self.object.Q0))
                vmax = max(max(self.object.Q0, key=max))
                vmin = min(min(self.object.Q0, key=min))
                self.im_Q0.set_clim(vmin=vmin)
                self.im_Q0.set_clim(vmax=vmax)
                self.figure_Q0.canvas.draw()

    def compute(self, time_interval, write_interval, solver='explicit_k(x)',
                verbose=True):
        """Compute the thermal process.

        Computes the system for `time_interval`, and writes into the
        `file_name` file every `write_interval` time steps. Two different
        solvers can be used: `'explicit_general'` and `'explicit_k(x)'`. If
        verbose = True, then the progress of the computation is shown.

        """
        # check the validity of inputs
        cond1 = isinstance(time_interval, float)
        cond1 = cond1 or isinstance(time_interval, int)
        cond2 = isinstance(write_interval, int)
        all_solvers = ['explicit_general', 'explicit_k(x)']
        cond3 = isinstance(solver, str) and solver in all_solvers
        cond4 = isinstance(verbose, bool)
        condition = cond1 and cond2 and cond3 and cond4
        if not condition:
            raise ValueError

        # number of time steps for the given timeInterval
        nt = int(time_interval / self.object.dt)

        # number of time steps counting from the last writing process
        nw = 0

        # computes
        for k in range(nt):

            # updates the time_passed
            self.time_passed = self.time_passed + self.object.dt

            # defines the material properties accoring to the state list
            for i in range(self.object.size[0]):
                for j in range(self.object.size[1]):
                    if self.object.state[i][j] is True:
                        ix = self.object.materials_index[i][j]
                        self.object.rho[i][j] = self.object.materials[ix].rhoa(
                            self.object.temperature[i][j][0])
                        self.object.Cp[i][j] = self.object.materials[ix].cpa(
                            self.object.temperature[i][j][0])
                        self.object.k[i][j] = self.object.materials[ix].ka(
                            self.object.temperature[i][j][0])
                    if self.object.state[i][j] is False:
                        ix = self.object.materials_index[i][j]
                        self.object.rho[i][j] = self.object.materials[ix].rho0(
                            self.object.temperature[i][j][0])
                        self.object.Cp[i][j] = self.object.materials[ix].cp0(
                            self.object.temperature[i][j][0])
                        self.object.k[i][j] = self.object.materials[ix].k0(
                            self.object.temperature[i][j][0])

            # SOLVERS
            # explicit k constant
            if solver == 'explicit_general':
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])
                value = solvers.explicit_general(self.object)
                self.object.temperature, self.object.lheat = value
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])
            # explicit k constant
            if solver == 'explicit_k(x)':
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])
                value = solvers.explicit_k(self.object)
                self.object.temperature, self.object.lheat = value
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])

            nw = nw + 1

            if self.object.file_name:
                if nw + 1 == write_interval or k == 0 or k == nt - 1:
                    line = '%f' % self.time_passed
                    for i in range(self.object.size[0]):
                        for j in range(self.object.size[1]):
                            new_line = ',%f' % self.object.temperature[i][j][1]
                            line = line + new_line
                    f = open(self.object.file_name, 'a')
                    f.write(line+'\n')
                    f.close()
            if self.draw:
                for drawing in self.draw:
                    if drawing == 'temperature':
                        if nw + 1 == write_interval or k == 0 or k == nt - 1:
                            temp = []
                            for i in range(self.object.size[0]):
                                temp.append([])
                                for j in range(self.object.size[1]):
                                    value = self.object.temperature[i][j][0]
                                    temp[-1].append(value)
                            try:
                                self.im.set_array(np.transpose(temp))
                                if not self.draw_scale:
                                    vmax = max(max(temp, key=max))
                                    vmin = min(min(temp, key=min))
                                    self.im.set_clim(vmin=vmin)
                                    self.im.set_clim(vmax=vmax)
                                self.figure.canvas.draw()
                            except Exception:
                                pass

            if nw == write_interval:
                nw = 0
                if verbose:
                    print('progress:', int(100*k/nt), '%', end="\r")
        if verbose:
            print('Finished simulation')



"""Solvers.

This submodule contains the solvers for the computation of the unidimensional
thermal processes.

"""

from .implicit_k import implicit_k
from .implicit_general import implicit_general
from .explicit_k import explicit_k
from .explicit_general import explicit_general

__all__ = ['implicit_k', 'implicit_general', 'explicit_k', 'explicit_general']



"""Contains the class object.

Used to create unidimensional thermal objects and apply some methods. It does
not include the compute method.

"""

from pathlib import Path
from ... import mats
import os
import copy


class Object:
    """Object class.

    This class creates a unidimensional thermal object. It includes two
    methods to apply and remove fields.

    """

    def __init__(self, amb_temperature, materials=('Cu',), borders=(1, 11),
                 materials_order=(0,), dx=0.01, dt=0.1, file_name=None,
                 boundaries=(0, 0), Q=[], Q0=[], initial_state=False,
                 materials_path=False):
        """Thermal object initialization.

        `amb_temperature` is the ambient temperature of the whole system.
        `materials` is the list of strings of all the used materials present in
        `material_path`. `borders` is a list of the points where there is a
        change of material. `materials_order` is a list of the materials list
        indexes that defines the material properties given by borders. `dx` and
        `dt` are the space and time steps, respectively. `file_name` is the
        file name where the temperature is saved. `boundaries` is a list of two
        entries that define the boundary condition for temperature. If 0 the
        boundary condition is insulation. `initial_state` is the initial state
        of the materials. True if there are an applied field and False if them
        field is absent. `materials_path` is absolute path of the materials
        database. If false, then the materials database is the standard
        heatrapy database. `Q` is a list of fixed heat source coefficient and
        `Q0` is a list of temperature dependent heat source coefficient.

        """
        # check the validity of inputs
        materials = tuple(materials)
        borders = tuple(borders)
        materials_order = tuple(materials_order)
        boundaries = tuple(boundaries)
        Q = list(Q)
        Q0 = list(Q0)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(materials, tuple)
        cond03 = isinstance(borders, tuple)
        cond04 = isinstance(materials_order, tuple)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dt, int) or isinstance(dt, float)
        cond07 = isinstance(file_name, str)
        cond07 = cond07 or (file_name is None)
        cond08 = isinstance(boundaries, tuple)
        cond09 = isinstance(Q, list)
        cond10 = isinstance(Q0, list)
        cond11 = isinstance(initial_state, bool)
        condition = cond01 and cond02 and cond03 and cond04 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        condition = condition and cond10 and cond11
        if not condition:
            raise ValueError

        # initial definitions
        self.borders = borders
        self.materials = [i for i in range(len(materials))]
        self.boundaries = boundaries
        self.amb_temperature = amb_temperature

        # loads the data for each material
        if not materials_path:
            for i in range(len(materials)):
                tadi = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + materials[i] + '/' +
                            'tadi.txt')
                tadd = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + materials[i] + '/' +
                            'tadd.txt')
                cpa = Path(os.path.dirname(os.path.realpath(__file__)) +
                           '/../../database/' + materials[i] + '/' + 'cpa.txt')
                cp0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                           '/../../database/' + materials[i] + '/' + 'cp0.txt')
                k0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + materials[i] + '/' + 'k0.txt')
                ka = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + materials[i] + '/' + 'ka.txt')
                rho0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + materials[i] + '/' +
                            'rho0.txt')
                rhoa = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + materials[i] + '/' +
                            'rhoa.txt')
                lheat0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                              '/../../database/' + materials[i] + '/' +
                              'lheat0.txt')
                lheata = Path(os.path.dirname(os.path.realpath(__file__)) +
                              '/../../database/' + materials[i] + '/' +
                              'lheata.txt')
                self.materials[i] = mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata)
        else:
            for i in range(len(materials)):
                tadi = Path(materials_path + materials[i] + '/' + 'tadi.txt')
                tadd = Path(materials_path + materials[i] + '/' + 'tadd.txt')
                cpa = Path(materials_path + materials[i] + '/' + 'cpa.txt')
                cp0 = Path(materials_path + materials[i] + '/' + 'cp0.txt')
                k0 = Path(materials_path + materials[i] + '/' + 'k0.txt')
                ka = Path(materials_path + materials[i] + '/' + 'ka.txt')
                rho0 = Path(materials_path + materials[i] + '/' + 'rho0.txt')
                rhoa = Path(materials_path + materials[i] + '/' + 'rhoa.txt')
                lheat0 = Path(materials_path + materials[i] + '/' +
                              'lheat0.txt')
                lheata = Path(
                    materials_path + materials[i] + '/' + 'lheata.txt'
                )
                self.materials[i] = mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata)

        # defines which are the properties of each material point
        self.materials_index = [None]
        for i in range(len(borders) - 1):
            for j in range(borders[i], borders[i + 1]):
                self.materials_index.append(materials_order[i])
        self.materials_index.append(None)

        # loads the inputs and creates the lists temperature, Cp, rho and k
        self.num_points = borders[-1] + 1
        self.file_name = file_name
        self.dx = dx
        self.dt = dt
        self.temperature = [[amb_temperature, amb_temperature]]
        self.latent_heat = [None]
        self.lheat = [None]
        self.Cp = [None]
        self.rho = [None]
        self.Q = [None]
        self.Q0 = [None]
        self.k = [self.materials[self.materials_index[1]].k0(amb_temperature)]
        for i in range(1, borders[-1]):
            self.temperature.append([amb_temperature, amb_temperature])
            self.rho.append(self.materials[self.materials_index[i]].rho0(
                amb_temperature))
            self.Cp.append(
                self.materials[self.materials_index[i]].cp0(amb_temperature))
            self.k.append(
                self.materials[self.materials_index[i]].k0(amb_temperature))
            self.Q.append(0.)
            self.Q0.append(0.)
            self.latent_heat.append(
                self.materials[self.materials_index[i]].lheat0()
            )
            self.lheat.append([])
            for lh in self.materials[self.materials_index[i]].lheat0():
                if self.temperature[i][1] < lh[0] and lh[1] > 0.:
                    self.lheat[-1].append([lh[0], 0.])
                if self.temperature[i][1] > lh[0] and lh[1] > 0.:
                    self.lheat[-1].append([lh[0], lh[1]])
                if self.temperature[i][1] < lh[0] and lh[1] < 0.:
                    self.lheat[-1].append([lh[0], -lh[1]])
                if self.temperature[i][1] > lh[0] and lh[1] < 0.:
                    self.lheat[-1].append([lh[0], 0.])
        self.temperature.append([amb_temperature, amb_temperature])
        self.rho.append(None)
        self.Cp.append(None)
        self.Q.append(None)
        self.Q0.append(None)
        self.lheat.append(None)
        self.latent_heat.append(None)
        self.k.append(
            self.materials[self.materials_index[-2]].k0(amb_temperature))

        if Q != []:
            self.Q = Q
        if Q0 != []:
            self.Q0 = Q0

        # initializes the state of each point (True if active and False if ...
        # unactive), the time, heat flux from the hot side and heat flux ...
        # from the cold side
        self.state = [initial_state for i in range(borders[-1] + 1)]
        self.time_passed = 0.

        self.Q_ref = copy.copy(self.Q)
        self.Q0_ref = copy.copy(self.Q0)

        if file_name:
            line = 'time(s)'
            for i in range(len(self.temperature)):
                line = line + ',T[' + str(i) + '] (K)'
            line = line + '\n'
            f = open(self.file_name, 'a')
            f.write(line)
            f.close()

    def activate(self, initial_point, final_point):
        """Activation of the material.

        activates a given space interval of the material, between the
        `initial_point` and `final_point`.

        """
        initial_point = int(initial_point)
        final_point = int(final_point)
        for i in range(initial_point, final_point):

            if self.state[i] is False:
                self.temperature[i][0] = self.temperature[i][0] + \
                    self.materials[self.materials_index[i]].tadi(
                        self.temperature[i][0])
                self.rho[i] = self.materials[self.materials_index[i]].rhoa(
                    self.temperature[i][0])
                self.Cp[i] = self.materials[self.materials_index[i]].cpa(
                    self.temperature[i][0])
                self.k[i] = self.materials[self.materials_index[i]].ka(
                    self.temperature[i][0])
                self.lheat[i] = []
                valh = self.materials[self.materials_index[i]].lheata()
                self.latent_heat[i] = valh
                for lh in self.latent_heat[i]:
                    if self.temperature[i][0] < lh[0] and lh[1] > 0.:
                        self.lheat[i].append([lh[0], 0.])
                    if self.temperature[i][0] > lh[0] and lh[1] > 0.:
                        self.lheat[i].append([lh[0], lh[1]])
                    if self.temperature[i][0] < lh[0] and lh[1] < 0.:
                        self.lheat[i].append([lh[0], -lh[1]])
                    if self.temperature[i][0] > lh[0] and lh[1] < 0.:
                        self.lheat[i].append([lh[0], 0.])
                self.state[i] = True

            else:
                message = 'point %f already activated' % float(i)
                print(message)

    def deactivate(self, initial_point, final_point):
        """Deactivation of the material.

        deactivates a given space interval of the material, between the
        `initial_point` and `final_point`.

        """
        initial_point = int(initial_point)
        final_point = int(final_point)
        for i in range(initial_point, final_point):

            if self.state[i] is True:
                self.temperature[i][0] = self.temperature[i][0] - \
                    self.materials[self.materials_index[i]].tadd(
                        self.temperature[i][0])
                self.rho[i] = self.materials[self.materials_index[i]].rho0(
                    self.temperature[i][0])
                self.Cp[i] = self.materials[self.materials_index[i]].cp0(
                    self.temperature[i][0])
                self.k[i] = self.materials[self.materials_index[i]].k0(
                    self.temperature[i][0])
                self.lheat[i] = []
                valh = self.materials[self.materials_index[i]].lheat0()
                self.latent_heat[i] = valh
                for lh in self.latent_heat[i]:
                    if self.temperature[i][0] < lh[0] and lh[1] > 0.:
                        self.lheat[i].append([lh[0], 0.])
                    if self.temperature[i][0] > lh[0] and lh[1] > 0.:
                        self.lheat[i].append([lh[0], lh[1]])
                    if self.temperature[i][0] < lh[0] and lh[1] < 0.:
                        self.lheat[i].append([lh[0], -lh[1]])
                    if self.temperature[i][0] > lh[0] and lh[1] < 0.:
                        self.lheat[i].append([lh[0], 0.])
                self.state[i] = False

            else:
                message = 'point %f already deactivated' % float(i)
                print(message)



"""Contains the classes system_objects and single_object.

Used to compute two-dimensional thermal objects

"""

import copy
from .. import solvers
from . import Object


class SystemObjects:
    """System_objects class.

    This class creates a system of 2D thermal objects, establishes contact
    between them and computes the respective thermal processes.

    """

    def __init__(self, number_objects=2, materials=('Cu', 'Cu'),
                 objects_length=((10, 10), (10, 10)), amb_temperature=293,
                 dx=0.01, dy=0.01, dt=0.1, file_name=None, initial_state=False,
                 boundaries=((0, 0, 0, 0), (0, 0, 0, 0)),
                 materials_path=False):
        """System object initialization.

        `number_objects` is the integer number of thermal objects. `materials`
        is the list of strings of all the used materials present in
        `material_path`. `amb_temperature` is the ambient temperature of the
        whole system. `object_length` is the list of thermal object lengths
        (tuple of spacial steps). `dx` and `dy` are the space steps along the
        x- and y-axis, respectively. dt is the time step. `file_name` is the
        file name where the temperature is saved. `boundaries` is a list of
        four entries that define the boundary condition for temperature (left,
        right, bottom, top). If 0, the boundary condition is insulation.
        `materials_path` is absolute path of the materials database. If false,
        then the materials database is the standard heatrapy database.

        """
        # check the validity of inputs
        materials = tuple(materials)
        objects_length = tuple(objects_length)
        boundaries = tuple(boundaries)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(materials, tuple)
        cond03 = isinstance(number_objects, int)
        cond04 = isinstance(objects_length, tuple)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dy, int) or isinstance(dy, float)
        cond07 = isinstance(dt, int) or isinstance(dt, float)
        cond08 = isinstance(file_name, str)
        cond08 = cond08 or (file_name is None)
        cond09 = isinstance(boundaries, tuple)
        cond10 = isinstance(initial_state, bool)
        condition = cond01 and cond02 and cond03 and cond04 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        condition = condition and cond10
        if not condition:
            raise ValueError

        # initial definitions
        self.objects = []
        file_name_obj = None
        for i in range(number_objects):

            if file_name:
                file_name_obj = file_name + '_' + str(i) + '.txt'

            self.objects.append(Object(amb_temperature, material=materials[i],
                                       dx=dx, dy=dy, dt=dt,
                                       file_name=file_name_obj,
                                       boundaries=boundaries[i], Q=[], Q0=[],
                                       initial_state=initial_state,
                                       materials_path=materials_path))
        self.contacts = set()
        self.boundaries = boundaries
        self.dt = dt
        self.q1 = 0.
        self.q2 = 0.

    def contact_filter(self, object_id):
        """Filter thermal contacts.

        Filter thermal contacts by `object_id`.

        """
        # check the validity of inputs
        condition = isinstance(object_id, int)
        if not condition:
            raise ValueError

        value = object_id
        filtered = [x for x in
                    self.contacts if (x[0][0] == value or x[1][0] == value)]
        return set(filtered)

    def contact_add(self, contact):
        """Add contact to self.contacts.

        `contact` is a thermal contact tuple of length 3, where the first and
        second entries correspond to tuples of the thermal objects and points
        (object_id, (x,y)), and the third entry is the heat transfer
        coefficient.

        """
        # check the validity of inputs
        if isinstance(contact, tuple):
            if len(contact) == 3:
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.contacts.add(contact)

    def contact_remove(self, object_one, object_two):
        """Contact removal.

        Removes all contacts between `object_one` id and `object_two` id.

        """
        # check the validity of inputs
        condition = isinstance(object_one, int)
        condition = condition and isinstance(object_two, int)
        if not condition:
            raise ValueError

        contact_list = list(self.contacts)
        for i in range(len(contact_list)):
            cond_1 = contact_list[i][0][0] == object_one
            cond_1 = cond_1 and contact_list[i][1][0] == object_two
            cond_2 = contact_list[i][0][0] == object_two
            cond_2 = cond_2 and contact_list[i][1][0] == object_one
            if cond_1 or cond_2:
                self.contacts.remove(contact_list[i])

    def change_boundaries(self, object_id, boundaries):
        """Change boundaries.

        Changes the `boundaries` variable of `object_id`.

        """
        # check the validity of inputs
        condition = isinstance(object_id, int)
        condition = condition and isinstance(boundaries, tuple)
        if condition:
            if len(boundaries) == 4:
                condition = True
            else:
                condition = False
        if not condition:
            raise ValueError

        self.objects[object_id].boundaries = boundaries

    def compute(self, time_interval, write_interval, solver='explicit_k(x)',
                verbose=True):
        """Compute the thermal process.

        Computes the system for `time_interval`, and writes into the
        `file_name` file every `write_interval` time steps. Two different
        solvers can be used: `'explicit_general'` and `'explicit_k(x)'`. If
        verbose = True, then the progress of the computation is shown.

        """
        # check the validity of inputs
        cond1 = isinstance(time_interval, float)
        cond1 = cond1 or isinstance(time_interval, int)
        cond2 = isinstance(write_interval, int)
        cond3 = isinstance(solver, str)
        cond4 = isinstance(verbose, bool)
        condition = cond1 and cond2 and cond3 and cond4
        if not condition:
            raise ValueError

        # number of time steps for the given time_interval
        nt = int(time_interval / self.dt)

        # number of time steps counting from the last writing process
        nw = 0

        # computes
        for k in range(nt):
            for obj in self.objects:
                obj.Q0 = copy.copy(obj.Q0_ref)

            for contact in self.contacts:
                in1_x = int(contact[1][1][0])
                in1_y = int(contact[1][1][1])
                in2_x = int(contact[0][1][0])
                in2_y = int(contact[0][1][1])
                td1 = self.objects[contact[1][0]].temperature[in1_x][in1_y][0]
                td2 = self.objects[contact[0][0]].temperature[in2_x][in2_y][0]
                heat_contact_1 = contact[2] * (td1 - td2)
                heat_contact_2 = contact[2] * (td2 - td1)
                self.objects[contact[0][0]].Q0[in2_x][in2_y] = heat_contact_1
                self.objects[contact[1][0]].Q0[in1_x][in2_y] = heat_contact_2

            for obj in self.objects:
                obj.time_passed = obj.time_passed + obj.dt

                # defines the material properties
                for i in range(obj.size[0]):
                    for j in range(obj.size[1]):
                        if obj.state[i][j] is True:
                            ind = obj.materials_index[i][j]
                            obj.rho[i][j] = obj.materials[ind].rhoa(
                                obj.temperature[i][j][0])
                            obj.Cp[i][j] = obj.materials[ind].cpa(
                                obj.temperature[i][j][0])
                            obj.k[i][j] = obj.materials[ind].ka(
                                obj.temperature[i][j][0])
                        if obj.state[i][j] is False:
                            ind = obj.materials_index[i][j]
                            obj.rho[i][j] = obj.materials[ind].rho0(
                                obj.temperature[i][j][0])
                            obj.Cp[i][j] = obj.materials[ind].cp0(
                                obj.temperature[i][j][0])
                            obj.k[i][j] = obj.materials[ind].k0(
                                obj.temperature[i][j][0])

                # explicit k constant
                if solver == 'explicit_general':
                    temp = []
                    for i in range(obj.size[0]):
                        temp.append([])
                        for j in range(obj.size[1]):
                            temp[-1].append(obj.temperature[i][j][0])
                    obj.temperature, obj.lheat = solvers.explicit_general(obj)
                    temp = []
                    for i in range(obj.size[0]):
                        temp.append([])
                        for j in range(obj.size[1]):
                            temp[-1].append(obj.temperature[i][j][0])

                # explicit k constant
                if solver == 'explicit_k(x)':
                    temp = []
                    for i in range(obj.size[0]):
                        temp.append([])
                        for j in range(obj.size[1]):
                            temp[-1].append(obj.temperature[i][j][0])
                    obj.temperature, obj.lheat = solvers.explicit_k(obj)
                    temp = []
                    for i in range(obj.size[0]):
                        temp.append([])
                        for j in range(obj.size[1]):
                            temp[-1].append(obj.temperature[i][j][0])

                if obj.file_name:
                    if nw + 1 == write_interval or k == 0 or k == nt - 1:
                        line = '%f' % obj.time_passed
                        for i in range(obj.size[0]):
                            for j in range(obj.size[1]):
                                new_line = ',%f' % obj.temperature[i][j][1]
                                line = line + new_line
                        f = open(obj.file_name, 'a')
                        f.write(line+'\n')
                        f.close()

            if nw == write_interval:
                nw = 0
                if verbose:
                    print('progress:', int(100*k/nt), '%', end='\r')
            else:
                nw = nw + 1

        if verbose:
            print('Finished simulation')



"""Contains the 2D explicit_k solver.

Used to compute two-dimensional thermal processes

"""

import copy


def explicit_k(obj):
    """explicit_k solver.

    Used to compute one time step of 2D systems with position dependent thermal
    conductivities.

    """
    x = copy.deepcopy(obj.temperature)
    temp_k = copy.deepcopy(obj.k)
    # bottom boundary
    x.append([])
    temp_k.append([])
    if obj.boundaries[1] == 0:
        for j in range(obj.size[1]):
            x[-1].append(obj.temperature[-1][j])
            temp_k[-1].append(obj.k[-1][j])
    else:
        for j in range(obj.size[1]):
            x[-1].append([obj.boundaries[1], obj.boundaries[1]])
            temp_k[-1].append(obj.k[-1][j])

    # top boundary
    x.insert(0, [])
    temp_k.insert(0, [])
    if obj.boundaries[0] == 0:
        for j in range(obj.size[1]):
            x[0].append(obj.temperature[0][j])
            temp_k[0].append(obj.k[0][j])
    else:
        for j in range(obj.size[1]):
            x[0].append([obj.boundaries[0], obj.boundaries[0]])
            temp_k[0].append(obj.k[0][j])

    # left boundary
    if obj.boundaries[2] == 0:
        for i in range(obj.size[0]+2):
            x[i].insert(0, x[i][0])
            temp_k[i].insert(0, temp_k[i][0])
    else:
        for i in range(obj.size[0]+2):
            x[i].insert(0, [obj.boundaries[2], obj.boundaries[2]])
            temp_k[i].insert(0, temp_k[i][0])

    # right boundary
    if obj.boundaries[3] == 0:
        for i in range(obj.size[0]+2):
            x[i].append(x[i][-1])
            temp_k[i].append(temp_k[i][-1])
    else:
        for i in range(obj.size[0]+2):
            x[i].append([obj.boundaries[3], obj.boundaries[3]])
            temp_k[i].append(temp_k[i][-1])

    final_x = copy.deepcopy(x)

    # computes
    for i in range(1, obj.size[0]+1):
        for j in range(1, obj.size[1]+1):
            # print(i,j)
            alpha = obj.dt / (2 * obj.rho[i-1][j-1] * obj.Cp[i-1][j-1] *
                              obj.dx * obj.dx)

            gamma = obj.dt / (2 * obj.rho[i-1][j-1] * obj.Cp[i-1][j-1] *
                              obj.dy * obj.dy)

            beta = obj.dt / (obj.rho[i-1][j-1] * obj.Cp[i-1][j-1])

            t_new = ((1 + beta * obj.Q[i-1][j-1]) * x[i][j][0] +
                     alpha * ((temp_k[i+1][j]+temp_k[i][j])*x[i+1][j][0] -
                              (temp_k[i-1][j] + 2 * temp_k[i][j] +
                               temp_k[i+1][j])*x[i][j][0] +
                              (temp_k[i-1][j]+temp_k[i][j])*x[i-1][j][0]) +
                     gamma * ((temp_k[i][j+1]+temp_k[i][j])*x[i][j+1][0] -
                              (temp_k[i][j-1] + 2 * temp_k[i][j] +
                               temp_k[i][j+1])*x[i][j][0] +
                              (temp_k[i][j-1]+temp_k[i][j])*x[i][j-1][0]) +
                     beta * (obj.Q0[i-1][j-1] - obj.Q[i-1][j-1] *
                             obj.amb_temperature))
            final_x[i][j][1] = t_new

    x = final_x
    # updates temperature for next iteration
    for i in range(len(x)):
        for j in range(len(x[1])):
            x[i][j][0] = x[i][j][1]

    # remove boundaries
    removed_temp = x.pop(0)
    removed_temp = x.pop()
    for i in range(len(x)):
        removed_temp = x[i].pop(0)
        removed_temp = x[i].pop()
    del removed_temp

    nx = []
    for i in range(obj.size[0]):
        nx.append([])
        for j in range(obj.size[1]):
            nx[-1].append(x[i][j][0])

    # print(x)
    # print('\n')
    # latent heat
    lheat = copy.copy(obj.lheat)
    for i in range(obj.size[0]):
        for j in range(obj.size[1]):
            k = 0
            for lh in obj.latent_heat[i][j]:
                temper = obj.temperature[i][j][0]
                value = nx[i][j] > lh[0]
                if value and temper <= lh[0] and lheat[i][j][k][1] != lh[1]:
                    en = (obj.Cp[i][j] * obj.rho[i][j] *
                          (nx[i][j] - obj.temperature[i][j][0]))
                    if en + lheat[i][j][k][1] >= lh[1]:
                        lheat[i][j][k][1] = lh[1]
                        energy_temp = lheat[i][j][k][1] + en - lh[1]
                        nx[i][j] = obj.temperature[i][j][0] + \
                            energy_temp / (obj.Cp[i][j] * obj.rho[i][j])
                    else:
                        lheat[i][j][k][1] += en
                        nx[i][j] = obj.temperature[i][j][0]
                value = nx[i][j] < lh[0]
                if value and temper >= lh[0] and lheat[i][j][k][1] != 0:
                    en = (obj.Cp[i][j] * obj.rho[i][j] *
                          (nx[i][j] - obj.temperature[i][j][0]))
                    if en + lheat[i][j][k][1] <= 0.:
                        lheat[i][j][k][1] = 0.
                        energy_temp = (en + lheat[i][j][k][1])
                        nx[i][j] = obj.temperature[i][j][0] + \
                            energy_temp / (obj.Cp[i][j] * obj.rho[i][j])
                    else:
                        lheat[i][j][k][1] += en
                        nx[i][j] = obj.temperature[i][j][0]
                k += 1

    y = copy.deepcopy(obj.temperature)

    # updates the temperature list
    for i in range(obj.size[0]):
        for j in range(obj.size[1]):
            y[i][j][1] = nx[i][j]
            y[i][j][0] = nx[i][j]

    return y, lheat



"""Contains the class system_objects.

Used to compute systems of thermal objects.

"""

import copy
from .. import solvers
from . import Object


class SystemObjects:
    """System_objects class.

    This class creates a system of unidimensional thermal objects, establishes
    contact between them and computes the respective thermal processes.

    """

    def __init__(self, number_objects=2, materials=('Cu', 'Cu'),
                 objects_length=(10, 10), amb_temperature=293, dx=0.01, dt=0.1,
                 file_name=None, initial_state=False,
                 boundaries=((2, 0), (3, 0)), materials_path=False):
        """System object initialization.

        `number_objects` is the integer number of thermal objects. `materials`
        is the list of strings of all the used materials present in
        `material_path`. `amb_temperature` is the ambient temperature of the
        whole system. `object_length` is the list of thermal object lengths
        (spacial steps). `dx` and `dt` are the space and time steps,
        respectively. `file_name` is the file name where the temperature is
        saved. `boundaries` is a list of tuples of length two that define each
        boundary condition for temperature. If 0 the boundary condition is
        insulation. `materials_path` is absolute path of the materials
        database. If false, then the materials database is the standard
        heatrapy database.

        """
        # check the validity of inputs
        materials = tuple(materials)
        objects_length = tuple(objects_length)
        boundaries = tuple(boundaries)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(materials, tuple)
        cond03 = isinstance(number_objects, int)
        cond04 = isinstance(objects_length, tuple)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dt, int) or isinstance(dt, float)
        cond07 = isinstance(file_name, str)
        cond07 = cond07 or (file_name is None)
        cond08 = isinstance(boundaries, tuple)
        cond09 = isinstance(initial_state, bool)
        condition = cond01 and cond02 and cond03 and cond04 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        if not condition:
            raise ValueError

        # initial definitions
        self.objects = []
        for i in range(number_objects):
            if file_name:
                file_name = file_name + '_' + str(i) + '.txt'

            self.objects.append(Object(amb_temperature,
                                materials=(materials[i],),
                                borders=(1, objects_length[i]+1),
                                materials_order=(0,), dx=dx, dt=dt,
                                file_name=file_name, boundaries=(0, 0),
                                Q=[], Q0=[], initial_state=initial_state,
                                materials_path=materials_path))

        self.contacts = set()
        self.boundaries = boundaries
        self.dt = dt
        self.q1 = 0.
        self.q2 = 0.

        for i in boundaries:
            if i[1] != 0:
                for j in range(len(self.objects[i[0]].temperature)):
                    self.objects[i[0]].temperature[j] = [i[1], i[1]]

    def contact_filter(self, object):
        """Filter self.contacts by thermal object id.

        object: thermal object id

        """
        # check the validity of inputs
        condition = object in range(len(self.objects))
        if not condition:
            raise ValueError

        filtered = [x for x in
                    self.contacts if (x[0][0] == object or x[1][0] == object)]
        return set(filtered)

    def contact_add(self, contact):
        """Add contact to self.contacts.

        The `contact` parameter is a tuple of length 3 (one element for thermal
        object A, one for thermal object B, and one for the heat transfer
        coefficient). Each thermal object element is a tuple of length 2 where
        the first element is the index of the thermal object and the second is
        the spatial point index.

        """
        # check the validity of inputs
        if isinstance(contact, list) or isinstance(contact, tuple):
            if len(contact) == 3:
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.contacts.add(contact)

    def contact_remove(self, object_one, object_two):
        """Contact removal.

        Removes all contacts between `object_one` id and `object_two` id.

        """
        # check the validity of inputs
        condition = isinstance(object_one, int)
        condition = condition and isinstance(object_two, int)
        if not condition:
            raise ValueError

        contact_list = list(self.contacts)
        for i in range(len(contact_list)):
            cond_1 = contact_list[i][0][0] == object_one
            cond_1 = cond_1 and contact_list[i][1][0] == object_two
            cond_2 = contact_list[i][0][0] == object_two
            cond_2 = cond_2 and contact_list[i][1][0] == object_one
            if cond_1 or cond_2:
                self.contacts.remove(contact_list[i])

    def change_boundaries(self, object_id, boundaries):
        """Change boundaries.

        Changes the `boundaries` of `object_id`.

        """
        # check the validity of inputs
        condition = isinstance(object_id, int)
        condition = condition and isinstance(boundaries, tuple)
        if condition:
            if len(boundaries) == 2:
                condition = True
            else:
                condition = False
        if not condition:
            raise ValueError

        self.objects[object_id].boundaries = boundaries

    def compute(self, time_interval, write_interval, solver='implicit_k(x)',
                verbose=True):
        """Compute the thermal process.

        Computes the system for `time_interval`, and writes into the
        `file_name` file every `write_interval` time steps. Four different
        solvers can be used: `'explicit_general'`, `'explicit_k(x)'`,
        `'implicit_general'`, and `'implicit_k(x)'`. If `verbose = True`, then
        the progress of the computation is shown.

        """
        # check the validity of inputs
        cond1 = isinstance(time_interval, float)
        cond1 = cond1 or isinstance(time_interval, int)
        cond2 = isinstance(write_interval, int)
        cond3 = isinstance(solver, str)
        cond4 = isinstance(verbose, bool)
        condition = cond1 and cond2 and cond3 and cond4
        if not condition:
            raise ValueError

        # number of time steps for the given timeInterval
        nt = int(time_interval / self.dt)

        # number of time steps counting from the last writing process
        nw = 0

        # computes
        for j in range(nt):
            for obj in self.objects:
                obj.Q0 = copy.copy(obj.Q0_ref)

            for contact in self.contacts:
                ind1 = int(contact[1][1])
                ind2 = int(contact[0][1])
                td1 = self.objects[contact[1][0]].temperature[ind1][0]
                td2 = self.objects[contact[0][0]].temperature[ind2][0]
                heat_contact_1 = contact[2] * (td1 - td2)
                heat_contact_2 = contact[2] * (td2 - td1)
                self.objects[contact[0][0]].Q0[ind2] = heat_contact_1
                self.objects[contact[1][0]].Q0[ind1] = heat_contact_2

            object_number = -1
            for obj in self.objects:
                object_number = object_number + 1
                obj.time_passed = obj.time_passed + obj.dt

                cond1 = object_number not in [
                    boundary[0] for boundary in self.boundaries
                ]
                if cond1 or (object_number, 0) in self.boundaries:

                    # defines the material properties
                    for i in range(1, obj.num_points - 1):
                        if obj.state[i] is True:
                            ind = obj.materials_index[i]
                            obj.rho[i] = obj.materials[ind].rhoa(
                                obj.temperature[i][0])
                            obj.Cp[i] = obj.materials[ind].cpa(
                                obj.temperature[i][0])
                            obj.k[i] = obj.materials[ind].ka(
                                obj.temperature[i][0])
                        if obj.state[i] is False:
                            ind = obj.materials_index[i]
                            obj.rho[i] = obj.materials[ind].rho0(
                                obj.temperature[i][0])
                            obj.Cp[i] = obj.materials[ind].cp0(
                                obj.temperature[i][0])
                            obj.k[i] = obj.materials[ind].k0(
                                obj.temperature[i][0])

                    # SOLVERS

                    # implicit k constant
                    if solver == 'implicit_general':
                        value = solvers.implicit_general(obj)
                        obj.temperature, obj.lheat = value

                    # implicit k dependent on x
                    if solver == 'implicit_k(x)':
                        obj.temperature, obj.lheat = solvers.implicit_k(obj)

                    # explicit k constant
                    if solver == 'explicit_general':
                        value = solvers.explicit_general(obj)
                        obj.temperature, obj.lheat = value

                    # explicit k dependent on x
                    if solver == 'explicit_k(x)':
                        obj.temperature, obj.lheat = solvers.explicit_k(obj)

                    # writes the temperature to file_name file ...
                    # if the number of time steps is verified
                    if obj.file_name:
                        if nw + 1 == write_interval or j == 0 or j == nt - 1:
                            line = '%f' % obj.time_passed
                            for i in obj.temperature:
                                new_line = ',%f' % i[1]
                                line = line + new_line
                            f = open(obj.file_name, 'a')
                            f.write(line+'\n')
                            f.close()

                else:

                    heat = [p*self.dt*obj.dx for p in obj.Q0 if p is not None]
                    heat = sum(heat)/(len(heat)*obj.dx)

                    if object_number == self.boundaries[0][0]:
                        self.q1 = self.q1 + heat
                        q = self.q1
                    else:
                        self.q2 = self.q2 + heat
                        q = self.q2

                    # writes the temperature to file_name file ...
                    # if the number of time steps is verified
                    if obj.file_name:
                        if nw + 1 == write_interval or j == 0 or j == nt - 1:
                            line = '%f' % obj.time_passed
                            for i in obj.temperature:
                                new_line = ',%f' % i[1]
                                line = line + new_line
                            new_line = ',%f' % q
                            line = line + new_line
                            f = open(obj.file_name, 'a')
                            f.write(line+'\n')
                            f.close()

            if nw == write_interval:
                nw = 0
                if verbose:
                    print('progress:', int(100*j/nt), '%', end='\r')
            else:
                nw = nw + 1

        if verbose:
            print('Finished simulation')



"""objects.

This submodule includes the two-dimensional object classes.

"""

from .object import Object
from .system import SystemObjects
from .single import SingleObject

__all__ = ['Object', 'SystemObjects', 'SingleObject']



"""Contains the class calmatpro.

Used to access the properties of materials from the database.

"""

import numpy as np


class CalMatPro:
    """CalMatPro class.

    Loads and gives physical properties of materials from the database

    """

    def __init__(self, name_tadi, name_tadd, name_cpa, name_cp0, name_k0,
                 name_ka, name_rho0, name_rhoa, name_lheat0, name_lheata):
        """Thermal properties from materials folder.

        `name_tadi` is the adiabatic temperature increase file and `name_tadd`
        is the adiabatic temperature decrease file. `name_cp0` is the specific
        heat file and `name_cpa` is the active specific heat file. `name_k0` is
        the thermal conductivity file and `name_ka` is the active thermal
        conductivity file. `name_rho0` is the density file and `name_rhoa` is
        the active density file. `name_lheat0` and `name_lheata` are the latent
        heat files when the material status is True or False, respectively.

        """
        # adiabatic temperature increase
        input = open(name_tadi, 'r')
        s = input.readlines()
        self.xTadi = []
        self.yTadi = []
        for line in s:
            pair_tad = line.split()
            self.xTadi.append(float(pair_tad[0]))
            self.yTadi.append(float(pair_tad[1]))
        input.close()

        # adiabatic temperature decrease
        input = open(name_tadd, 'r')
        s = input.readlines()
        self.xTadd = []
        self.yTadd = []
        for line in s:
            pair_tad = line.split()
            self.xTadd.append(float(pair_tad[0]))
            self.yTadd.append(float(pair_tad[1]))
        input.close()

        # specific heat
        input = open(name_cpa, 'r')
        s = input.readlines()
        self.xCpA = []
        self.yCpA = []
        for line in s:
            pair_tad = line.split()
            self.xCpA.append(float(pair_tad[0]))
            self.yCpA.append(float(pair_tad[1]))
        input.close()

        # active specific heat
        input = open(name_cp0, 'r')
        s = input.readlines()
        self.xCp0 = []
        self.yCp0 = []
        for line in s:
            pair_tad = line.split()
            self.xCp0.append(float(pair_tad[0]))
            self.yCp0.append(float(pair_tad[1]))
        input.close()

        # thermal conductivity
        input = open(name_k0, 'r')
        s = input.readlines()
        self.xk0 = []
        self.yk0 = []
        for line in s:
            pair_tad = line.split()
            self.xk0.append(float(pair_tad[0]))
            self.yk0.append(float(pair_tad[1]))
        input.close()

        # active thermal conductivity
        input = open(name_ka, 'r')
        s = input.readlines()
        self.xkA = []
        self.ykA = []
        for line in s:
            pair_tad = line.split()
            self.xkA.append(float(pair_tad[0]))
            self.ykA.append(float(pair_tad[1]))
        input.close()

        # density
        input = open(name_rho0, 'r')
        s = input.readlines()
        self.xrho0 = []
        self.yrho0 = []
        for line in s:
            pair_tad = line.split()
            self.xrho0.append(float(pair_tad[0]))
            self.yrho0.append(float(pair_tad[1]))
        input.close()

        # active density
        input = open(name_rhoa, 'r')
        s = input.readlines()
        self.xrhoA = []
        self.yrhoA = []
        for line in s:
            pair_tad = line.split()
            self.xrhoA.append(float(pair_tad[0]))
            self.yrhoA.append(float(pair_tad[1]))
        input.close()

        # latent heat
        input = open(name_lheat0, 'r')
        s = input.readlines()
        self.latent_heat0 = []
        for line in s:
            latent_values = line.split()
            self.latent_heat0.append(
                (float(latent_values[0]), float(latent_values[1]))
            )
        input.close()

        # active latent heat
        input = open(name_lheata, 'r')
        s = input.readlines()
        self.latent_heatA = []
        for line in s:
            latent_values = line.split()
            self.latent_heatA.append(
                (float(latent_values[0]), float(latent_values[1]))
            )
        input.close()

    def tadi(self, temperature):
        """Adiabatic temperature increase.

        Gives the adiabatic temperature increase for a given temperature

        """
        return np.interp(temperature, self.xTadi, self.yTadi)

    def tadd(self, temperature):
        """Adiabatic temperature decrease.

        Gives the adiabatic temperature decrease for a given temperature

        """
        return np.interp(temperature, self.xTadd, self.yTadd)

    def cpa(self, temperature):
        """Active specific heat.

        Gives the active specific heat for a given temperature

        """
        return np.interp(temperature, self.xCpA, self.yCpA)

    def cp0(self, temperature):
        """Inactive specific heat.

        Gives the specific heat for a given temperature

        """
        return np.interp(temperature, self.xCp0, self.yCp0)

    def k0(self, temperature):
        """Inactive thermal conductivity.

        Gives the thermal conductivity for a given temperature

        """
        return np.interp(temperature, self.xk0, self.yk0)

    def ka(self, temperature):
        """Active thermal conductivity.

        Gives the active thermal conductivity for a given temperature

        """
        return np.interp(temperature, self.xkA, self.ykA)

    def rho0(self, temperature):
        """Inactive density.

        Gives the density for a given temperature

        """
        return np.interp(temperature, self.xrho0, self.yrho0)

    def rhoa(self, temperature):
        """Active density.

        Gives the active density for a given temperature

        """
        return np.interp(temperature, self.xrhoA, self.yrhoA)

    def lheat0(self):
        """Inactive latent heat.

        Gives the latent heat list

        """
        return self.latent_heat0

    def lheata(self):
        """Active latent heat.

        Gives the active latent heat list

        """
        return self.latent_heatA



"""Contains the 1D explicit_general solver.

Used to compute unidimensional thermal processes

"""

import copy


def explicit_general(obj):
    """explicit_general solver.

    Used to compute one time step of 1D systems with fixed thermal
    conductivity.

    """
    x = copy.deepcopy(obj.temperature)

    # computes
    for i in range(1, obj.num_points - 1):

        alpha = obj.dt * \
            obj.k[i] / (obj.rho[i] * obj.Cp[i] *
                        obj.dx * obj.dx)
        beta = obj.dt / (obj.rho[i] * obj.Cp[i])

        t_new = ((1 + beta * obj.Q[i]) * obj.temperature[i][0] +
                 alpha * (obj.temperature[i - 1][0] - 2 *
                          obj.temperature[i][0] + obj.temperature[i + 1][0]) +
                 beta * (obj.Q0[i] - obj.Q[i] * obj.amb_temperature))
        x[i][1] = t_new

    # left boundary for next time step
    if obj.boundaries[0] == 0:
        x[0][1] = obj.temperature[1][1]
    else:
        x[0][1] = obj.boundaries[0]

    # right boundary for next time step
    if obj.boundaries[1] == 0:
        x[obj.num_points - 1][1] = obj.temperature[obj.num_points - 2][1]
    else:
        x[obj.num_points - 1][1] = obj.boundaries[1]

    # updates temperature for next iteration
    for i in range(0, obj.num_points):
        x[i][0] = x[i][1]

    nx = []
    for i in range(len(x)):
        nx.append(x[i][0])

    # latent heat
    lheat = copy.copy(obj.lheat)
    for i in range(1, obj.num_points - 1):
        j = 0
        for lh in obj.latent_heat[i]:
            temper = obj.temperature[i][0]
            if nx[i] > lh[0] and temper <= lh[0] and lheat[i][j][1] != lh[1]:
                en = obj.Cp[i] * obj.rho[i] * (nx[i] - obj.temperature[i][0])
                if en + lheat[i][j][1] >= lh[1]:
                    lheat[i][j][1] = lh[1]
                    energy_temp = lheat[i][j][1] + en - lh[1]
                    nx[i] = obj.temperature[i][0] + \
                        energy_temp / (obj.Cp[i] * obj.rho[i])
                else:
                    lheat[i][j][1] += en
                    nx[i] = obj.temperature[i][0]
            if nx[i] < lh[0] and temper >= lh[0] and lheat[i][j][1] != 0:
                en = obj.Cp[i] * obj.rho[i] * (nx[i] - obj.temperature[i][0])
                if en + lheat[i][j][1] <= 0.:
                    lheat[i][j][1] = 0.
                    energy_temp = (en + lheat[i][j][1])
                    nx[i] = obj.temperature[i][0] + \
                        energy_temp / (obj.Cp[i] * obj.rho[i])
                else:
                    lheat[i][j][1] += en
                    nx[i] = obj.temperature[i][0]
            j += 1

    y = copy.deepcopy(obj.temperature)

    # updates the temperature list
    for i in range(obj.num_points):
        y[i][1] = nx[i]
        y[i][0] = nx[i]

    return y, lheat



# -*- coding: utf-8 -*-
"""heatrapy: Library for simulating heat transfer processes

Computes heat transfer processes in solids for 1D and 2D models.

"""

from setuptools import setup

description = 'This package is a module for simulating dynamic 1D and 2D heat '
description += 'transfer processes by using the finite difference method. The '
description += 'packages is based on thermal objects. Two different '
description += 'approaches can be used: single thermal object and a system of '
description += 'thermal objects that can contact with each other. The heatrapy'
description += 'package includes the incorporation of phase transitions.'


setup(name='heatrapy',
      version='2.0.4',
      description='Library for simulating 1D and 2D heat transfer processes',
      long_description=description,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.7',
          'Topic :: Scientific/Engineering :: Physics',
      ],
      url='https://github.com/djsilva99/heatrapy',
      author='Daniel Silva',
      author_email='djsilva@gmx.com',
      license='MIT',
      packages=['heatrapy'],
      install_requires=[
          'numpy>=1.20.1', 'matplotlib==3.3.4'
      ],
      include_package_data=True,
      zip_safe=False)



"""Contains the class object.

Used to create two-dimensional thermal objects, and apply some methods. It does
not include the compute method.

"""

from pathlib import Path
from ... import mats
import os
import copy
import numpy as np


class Object:
    """Object class.

    This class creates a two-dimensional thermal object. It includes two
    methods to apply and remove fields.

    """

    def __init__(self, amb_temperature, material='Cu', dx=0.01, dy=0.01,
                 dt=0.1, size=(10, 10), file_name=None,
                 boundaries=(0, 0, 0, 0), Q=[], Q0=[], initial_state=False,
                 materials_path=False):
        """Thermal object initialization.

        `amb_temperature` is the ambient temperature of the whole system.
        `materials` is the background material present in `material_path`.
        `dx`, `dy` are the space steps along the x- and y-axis, respectively.
        `dt` is the time step. `file_name` is the file name where the
        temperature is saved. `boundaries` is a list of four entries that
        define the boundary condition for temperature (left, right, bottom,
        top). If 0 the boundary condition is insulation. `initial_state` is the
        initial state of the materials. True if there are an applied field and
        False if them field is absent. `materials_path` is absolute path of the
        materials database. If false, then the materials database is the
        standard heatrapy database. `Q` is a list of fixed heat source
        coefficient and `Q0` is a list of temperature dependent heat source
        coefficient.

        """
        # check the validity of inputs
        boundaries = tuple(boundaries)
        Q = list(Q)
        Q0 = list(Q0)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dt, int) or isinstance(dt, float)
        cond07 = isinstance(file_name, str)
        cond07 = cond07 or (file_name is None)
        cond08 = isinstance(boundaries, tuple)
        cond09 = isinstance(Q, list)
        cond10 = isinstance(Q0, list)
        cond11 = isinstance(initial_state, bool)
        condition = cond01 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        condition = condition and cond10 and cond11
        if not condition:
            raise ValueError

        self.materials = [material]
        self.materials_name = [material]
        self.boundaries = boundaries
        self.amb_temperature = amb_temperature

        if materials_path is False:
            tadi = Path(os.path.dirname(os.path.realpath(__file__)) +
                        '/../../database/' + self.materials[0] + '/' +
                        'tadi.txt')
            tadd = Path(os.path.dirname(os.path.realpath(__file__)) +
                        '/../../database/' + self.materials[0] + '/' +
                        'tadd.txt')
            cpa = Path(os.path.dirname(os.path.realpath(__file__)) +
                       '/../../database/' + self.materials[0] + '/' +
                       'cpa.txt')
            cp0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                       '/../../database/' + self.materials[0] + '/' +
                       'cp0.txt')
            k0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                      '/../../database/' + self.materials[0] + '/' + 'k0.txt')
            ka = Path(os.path.dirname(os.path.realpath(__file__)) +
                      '/../../database/' + self.materials[0] + '/' + 'ka.txt')
            rho0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                        '/../../database/' + self.materials[0] + '/' +
                        'rho0.txt')
            rhoa = Path(os.path.dirname(os.path.realpath(__file__)) +
                        '/../../database/' + self.materials[0] + '/' +
                        'rhoa.txt')
            lheat0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + self.materials[0] + '/' +
                          'lheat0.txt')
            lheata = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + self.materials[0] + '/' +
                          'lheata.txt')
            self.materials[0] = mats.CalMatPro(
                tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata)
        else:
            tadi = Path(materials_path + self.materials[0] + '/' + 'tadi.txt')
            tadd = Path(materials_path + self.materials[0] + '/' + 'tadd.txt')
            cpa = Path(materials_path + self.materials[0] + '/' + 'cpa.txt')
            cp0 = Path(materials_path + self.materials[0] + '/' + 'cp0.txt')
            k0 = Path(materials_path + self.materials[0] + '/' + 'k0.txt')
            ka = Path(materials_path + self.materials[0] + '/' + 'ka.txt')
            rho0 = Path(materials_path + self.materials[0] + '/' + 'rho0.txt')
            rhoa = Path(materials_path + self.materials[0] + '/' + 'rhoa.txt')
            lheat0 = Path(materials_path + self.materials[0] + '/' +
                          'lheat0.txt')
            lheata = Path(materials_path + self.materials[0] + '/' +
                          'lheata.txt')
            self.materials[0] = mats.CalMatPro(
                tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata)

        self.materials_index = [None]
        self.size = size
        self.file_name = file_name
        self.dx = dx
        self.dy = dy
        self.dt = dt
        self.temperature = []
        self.latent_heat = []
        self.lheat = []
        self.Cp = []
        self.rho = []
        self.Q = []
        self.Q0 = []
        self.k = []
        self.materials_index = []
        self.state = []
        for i in range(self.size[0]):
            self.state.append([])
            self.materials_index.append([])
            self.temperature.append([])
            self.latent_heat.append([])
            self.lheat.append([])
            self.Cp.append([])
            self.rho.append([])
            self.Q.append([0. for i in range(self.size[1])])
            self.Q0.append([0. for i in range(self.size[1])])
            self.k.append([])
            for j in range(self.size[1]):
                self.materials_index[-1].append(0)
                self.temperature[-1].append([amb_temperature, amb_temperature])
                self.state[-1].append(initial_state)
                if initial_state:
                    value = self.materials[self.materials_index[i][j]]
                    self.Cp[-1].append(value.cpa(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.rho[-1].append(value.rhoa(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.k[-1].append(value.ka(self.amb_temperature))
                    self.latent_heat[-1].append(
                        self.materials[self.materials_index[i][j]].lheata()
                    )
                    self.lheat[-1].append([])
                    value = self.materials[self.materials_index[i][j]]
                    for lh in value.lheata():
                        if self.temperature[i][j][1] < lh[0] and lh[1] > 0.:
                            self.lheat[-1][-1].append([lh[0], 0.])
                        if self.temperature[i][j][1] > lh[0] and lh[1] > 0.:
                            self.lheat[-1][-1].append([lh[0], lh[1]])
                        if self.temperature[i][j][1] < lh[0] and lh[1] < 0.:
                            self.lheat[-1][-1].append([lh[0], -lh[1]])
                        if self.temperature[i][j][1] > lh[0] and lh[1] < 0.:
                            self.lheat[-1][-1].append([lh[0], 0.])
                else:
                    value = self.materials[self.materials_index[i][j]]
                    self.Cp[-1].append(value.cp0(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.rho[-1].append(value.rho0(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.k[-1].append(value.k0(self.amb_temperature))
                    self.latent_heat[-1].append(
                        self.materials[self.materials_index[i][j]].lheat0()
                    )
                    self.lheat[-1].append([])
                    value = self.materials[self.materials_index[i][j]]
                    for lh in value.lheat0():
                        if self.temperature[i][j][1] < lh[0] and lh[1] > 0.:
                            self.lheat[-1][-1].append([lh[0], 0.])
                        if self.temperature[i][j][1] > lh[0] and lh[1] > 0.:
                            self.lheat[-1][-1].append([lh[0], lh[1]])
                        if self.temperature[i][j][1] < lh[0] and lh[1] < 0.:
                            self.lheat[-1][-1].append([lh[0], -lh[1]])
                        if self.temperature[i][j][1] > lh[0] and lh[1] < 0.:
                            self.lheat[-1][-1].append([lh[0], 0.])

        if Q != []:
            self.Q = Q
        if Q0 != []:
            self.Q0 = Q0

        self.time_passed = 0.

        self.Q_ref = copy.copy(self.Q)
        self.Q0_ref = copy.copy(self.Q0)

        if file_name:
            line = 'time(s)'
            for i in range(size[0]):
                for j in range(size[1]):
                    line = line + ',T[' + str(i) + '][' + str(j) + '] (K)'
            line = line + '\n'
            f = open(self.file_name, 'a')
            f.write(line)
            f.close()

    def activate(self, initial_point, final_point, shape='square'):
        """Activation of the material.

        Activates a given piece of material. If `shape` is `'square'`, then the
        `initial_point` is the tuple (x,y) of the bottom left point and the
        `final_point` is the tuple (x,y) of the top right point. If the shape
        is `'circle'`, the `initial_point` is the tuple (x,y) of the center of
        the circle and `final_point` is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        if shape == 'square':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            final_point_x = int(final_point[0])
            final_point_y = int(final_point[1])
            for i in range(initial_point_x, final_point_x):
                for j in range(initial_point_y, final_point_y):
                    if self.state[i][j] is False:
                        value = self.temperature[i][j][0]
                        self.temperature[i][j][0] = value + \
                            self.materials[self.materials_index[i][j]].tadi(
                                self.temperature[i][j][0])
                        value = self.materials_index[i][j]
                        self.rho[i][j] = self.materials[value].rhoa(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[value].cpa(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[value].ka(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[value].lheata()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                        self.state[i][j] = True
                    else:
                        message = 'point ({:d},{:d})'.format(i, j)
                        message = message + ' already activated'
                        print(message)

        if shape == 'circle':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            radius = int(final_point)
            for i in range(self.size[0]):
                for j in range(self.size[1]):
                    value = (i-initial_point_x)**2+(j-initial_point_y)**2
                    length = np.sqrt(value)
                    if length <= radius:
                        if self.state[i][j] is False:
                            value = self.temperature[i][j][0]
                            index_value = self.materials_index[i][j]
                            self.temperature[i][j][0] = value + \
                                self.materials[index_value].tadi(
                                    self.temperature[i][j][0])
                            self.rho[i][j] = self.materials[index_value].rhoa(
                                self.temperature[i][j][0])
                            self.Cp[i][j] = self.materials[index_value].cpa(
                                self.temperature[i][j][0])
                            self.k[i][j] = self.materials[index_value].ka(
                                self.temperature[i][j][0])
                            self.lheat[i][j] = []
                            valh = self.materials[index_value].lheata()
                            self.latent_heat[i][j] = valh
                            for lh in self.latent_heat[i][j]:
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], lh[1]])
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], -lh[1]])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                            self.state[i][j] = True
                        else:
                            message = 'point ({:d},{:d})'.format(i, j)
                            message = message + ' already activated'
                            print(message)

    def deactivate(self, initial_point, final_point, shape='square'):
        """Deactivation of the material.

        Deactivates a given piece of material. If `shape` is `'square'`, then
        the `initial_point` is the tuple (x,y) of the bottom left point and the
        `final_point` is the tuple (x,y) of the top right point. If the shape
        is `'circle'`, the `initial_point` is the tuple (x,y) of the center of
        the circle and `final_point` is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        if shape == 'square':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            final_point_x = int(final_point[0])
            final_point_y = int(final_point[1])
            for i in range(initial_point_x, final_point_x):
                for j in range(initial_point_y, final_point_y):
                    if self.state[i][j] is True:
                        value = self.temperature[i][j][0]
                        self.temperature[i][j][0] = value - \
                            self.materials[self.materials_index[i][j]].tadd(
                                self.temperature[i][j][0])
                        value = self.materials_index[i][j]
                        self.rho[i][j] = self.materials[value].rho0(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[value].cp0(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[value].k0(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[value].lheat0()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                        self.state[i][j] = False
                    else:
                        message = 'point ({:d},{:d})'.format(i, j)
                        message = message + ' already deactivated'
                        print(message)

        if shape == 'circle':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            radius = int(final_point)
            for i in range(self.size[0]):
                for j in range(self.size[1]):
                    value = (i-initial_point_x)**2+(j-initial_point_y)**2
                    length = np.sqrt(value)
                    if length <= radius:
                        if self.state[i][j] is False:
                            value = self.temperature[i][j][0]
                            index_value = self.materials_index[i][j]
                            self.temperature[i][j][0] = value - \
                                self.materials[index_value].tadd(
                                    self.temperature[i][j][0])
                            self.rho[i][j] = self.materials[index_value].rho0(
                                self.temperature[i][j][0])
                            self.Cp[i][j] = self.materials[index_value].cp0(
                                self.temperature[i][j][0])
                            self.k[i][j] = self.materials[index_value].k0(
                                self.temperature[i][j][0])
                            self.lheat[i][j] = []
                            valh = self.materials[index_value].lheat0()
                            self.latent_heat[i][j] = valh
                            for lh in self.latent_heat[i][j]:
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], lh[1]])
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], -lh[1]])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                            self.state[i][j] = False
                        else:
                            message = 'point ({:d},{:d})'.format(i, j)
                            message = message + ' already deactivated'
                            print(message)

    def square(self, material='Gd', initial_point=(3, 3), length=(3, 3),
               state=False, materials_path=False):
        """Material adding with rectangle shape.

        Adds a new material with a rectangle shape, where `initial_point` is
        the bottom left (x,y) tuple, `length` is the length along the two axis,
        state is the initial state of the material and `materials_path` is the
        absolute path of the materials database.

        """
        # check the validity of inputs
        value = isinstance(initial_point, tuple)
        if value and isinstance(length, tuple):
            cond1 = len(initial_point) == 2
            cond1 = cond1 and len(length) == 2
        else:
            cond1 = False
        cond2 = isinstance(material, str)
        cond3 = isinstance(state, bool)
        cond4 = isinstance(materials_path, str) or materials_path is False
        if not cond1 and cond2 and cond3 and cond4:
            raise ValueError

        initial_point_x = int(initial_point[0])
        initial_point_y = int(initial_point[1])
        final_point_x = initial_point_x + int(length[0])
        final_point_y = initial_point_y + int(length[1])
        if material in self.materials_name:
            index = self.materials_name.index(material)
        else:
            index = len(self.materials)
            self.materials_name.append(material)
            if materials_path is False:
                value = self.materials_name[index]
                tadi = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'tadi.txt')
                tadd = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'tadd.txt')
                cpa = Path(os.path.dirname(os.path.realpath(__file__)) +
                           '/../../database/' + value + '/' + 'cpa.txt')
                cp0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                           '/../../database/' + value + '/' + 'cp0.txt')
                k0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + value + '/' + 'k0.txt')
                ka = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + value + '/' + 'ka.txt')
                rho0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'rho0.txt')
                rhoa = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'rhoa.txt')
                lheat0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                              '/../../database/' + value + '/' + 'lheat0.txt')
                lheata = Path(os.path.dirname(os.path.realpath(__file__)) +
                              '/../../database/' + value + '/' + 'lheata.txt')
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))
            else:
                value = self.materials_name[index]
                tadi = Path(materials_path + value + '/' + 'tadi.txt')
                tadd = Path(materials_path + value + '/' + 'tadd.txt')
                cpa = Path(materials_path + value + '/' + 'cpa.txt')
                cp0 = Path(materials_path + value + '/' + 'cp0.txt')
                k0 = Path(materials_path + value + '/' + 'k0.txt')
                ka = Path(materials_path + value + '/' + 'ka.txt')
                rho0 = Path(materials_path + value + '/' + 'rho0.txt')
                rhoa = Path(materials_path + value + '/' + 'rhoa.txt')
                lheat0 = Path(materials_path + value + '/' + 'lheat0.txt')
                lheata = Path(materials_path + value + '/' + 'lheata.txt')
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))

        for i in range(initial_point_x, final_point_x):
            for j in range(initial_point_y, final_point_y):
                if state is False:
                    self.state[i][j] = False
                    self.materials_index[i][j] = index
                    self.rho[i][j] = self.materials[index].rho0(
                        self.temperature[i][j][0])
                    self.Cp[i][j] = self.materials[index].cp0(
                        self.temperature[i][j][0])
                    self.k[i][j] = self.materials[index].k0(
                        self.temperature[i][j][0])
                    self.lheat[i][j] = []
                    valh = self.materials[index].lheat0()
                    self.latent_heat[i][j] = valh
                    for lh in self.latent_heat[i][j]:
                        if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], 0.])
                        if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], lh[1]])
                        if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], -lh[1]])
                        if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], 0.])

                else:
                    self.state[i][j] = True
                    self.materials_index[i][j] = index
                    self.rho[i][j] = self.materials[index].rhoa(
                        self.temperature[i][j][0])
                    self.Cp[i][j] = self.materials[index].cpa(
                        self.temperature[i][j][0])
                    self.k[i][j] = self.materials[index].ka(
                        self.temperature[i][j][0])
                    self.lheat[i][j] = []
                    valh = self.materials[index].lheata()
                    self.latent_heat[i][j] = valh
                    for lh in self.latent_heat[i][j]:
                        if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], 0.])
                        if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], lh[1]])
                        if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], -lh[1]])
                        if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], 0.])

    def circle(self, material='Gd', initial_point=(3, 3), radius=3,
               state=False, materials_path=False):
        """Material adding with circle shape.

        Adds a new material with a circle shape, where `initial_point` is the
        (x,y) tuple of the center of the circle, `radius` is the radius of the
        circle, state is the initial state of the material and `materials_path`
        is the absolute path of the materials database.

        """
        # check the validity of inputs
        if isinstance(initial_point, tuple):
            cond1 = len(initial_point) == 2
        else:
            cond1 = False
        cond2 = isinstance(radius, int) or isinstance(radius, float)
        cond3 = isinstance(material, str)
        cond4 = isinstance(state, bool)
        cond5 = isinstance(materials_path, str) or materials_path is False
        if not cond1 and cond2 and cond3 and cond4 and cond5:
            raise ValueError

        initial_point_x = int(initial_point[0])
        initial_point_y = int(initial_point[1])
        if material in self.materials_name:
            index = self.materials_name.index(material)
        else:
            index = len(self.materials)
            self.materials_name.append(material)
            if materials_path is False:
                value = self.materials_name[index]
                tadi = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'tadi.txt')
                tadd = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'tadd.txt')
                cpa = Path(os.path.dirname(os.path.realpath(__file__)) +
                           '/../../database/' + value + '/' + 'cpa.txt')
                cp0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                           '/../../database/' + value + '/' + 'cp0.txt')
                k0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + value + '/' + 'k0.txt')
                ka = Path(os.path.dirname(os.path.realpath(__file__)) +
                          '/../../database/' + value + '/' + 'ka.txt')
                rho0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'rho0.txt')
                rhoa = Path(os.path.dirname(os.path.realpath(__file__)) +
                            '/../../database/' + value + '/' + 'rhoa.txt')
                lheat0 = Path(os.path.dirname(os.path.realpath(__file__)) +
                              '/../../database/' + value + '/' + 'lheat0.txt')
                lheata = Path(os.path.dirname(os.path.realpath(__file__)) +
                              '/../../database/' + value + '/' + 'lheata.txt')
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))
            else:
                value = self.materials_name[index]
                tadi = Path(materials_path + value + '/' + 'tadi.txt')
                tadd = Path(materials_path + value + '/' + 'tadd.txt')
                cpa = Path(materials_path + value + '/' + 'cpa.txt')
                cp0 = Path(materials_path + value + '/' + 'cp0.txt')
                k0 = Path(materials_path + value + '/' + 'k0.txt')
                ka = Path(materials_path + value + '/' + 'ka.txt')
                rho0 = Path(materials_path + value + '/' + 'rho0.txt')
                rhoa = Path(materials_path + value + '/' + 'rhoa.txt')
                lheat0 = Path(materials_path + value + '/' + 'lheat0.txt')
                lheata = Path(materials_path + value + '/' + 'lheata.txt')
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))

        for i in range(self.size[0]):
            for j in range(self.size[1]):
                length = np.sqrt((i-initial_point_x)**2+(j-initial_point_y)**2)
                if length <= radius:
                    if state is False:
                        self.state[i][j] = False
                        self.materials_index[i][j] = index
                        self.rho[i][j] = self.materials[index].rho0(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[index].cp0(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[index].k0(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[index].lheat0()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], 0.])
                            if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], 0.])

                    else:
                        self.state[i][j] = True
                        self.materials_index[i][j] = index
                        self.rho[i][j] = self.materials[index].rhoa(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[index].cpa(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[index].ka(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[index].lheata()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], 0.])
                            if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], 0.])

    def power_add(self, initial_point, final_point, power, shape='square',
                  power_type='Q'):
        """Power adding.

        Adds a power matrix to the thermal object. If `shape` is `'square'`,
        then the `initial_point` is the tuple (x,y) of the bottom left point
        and the `final_point` is the tuple (x,y) of the top right point. If the
        `shape` is `'circle'`, the `initial_point` is the tuple (x,y) of the
        center of the circle and `final_point` is its radius. `power` is the
        value of the power to add, and `power_type` is the type of power to be
        introduced, which has the value `'Q'` if it is temperature dependent
        and `'Q0'` if it is temperature independent.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    cond1 = len(initial_point) == 2
                    cond1 = cond1 and len(final_point) == 2
                else:
                    cond1 = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    cond1 = len(initial_point) == 2
                else:
                    cond1 = False
            else:
                cond1 = False
        else:
            cond1 = False
        cond2 = isinstance(power, int) or isinstance(power, float)
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        if not cond1 and cond2 and cond3:
            raise ValueError

        if power_type == 'Q':
            if shape == 'square':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                final_point_x = int(final_point[0])
                final_point_y = int(final_point[1])
                for i in range(initial_point_x, final_point_x):
                    for j in range(initial_point_y, final_point_y):
                        self.Q[i][j] = power
            if shape == 'circle':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                radius = int(final_point)
                for i in range(self.size[0]):
                    for j in range(self.size[1]):
                        value = (i-initial_point_x)**2+(j-initial_point_y)**2
                        length = np.sqrt(value)
                        if length <= radius:
                            self.Q[i][j] = power

        if power_type == 'Q0':
            if shape == 'square':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                final_point_x = int(final_point[0])
                final_point_y = int(final_point[1])
                for i in range(initial_point_x, final_point_x):
                    for j in range(initial_point_y, final_point_y):
                        self.Q0[i][j] = power
            if shape == 'circle':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                radius = int(final_point)
                for i in range(self.size[0]):
                    for j in range(self.size[1]):
                        value = (i-initial_point_x)**2+(j-initial_point_y)**2
                        length = np.sqrt(value)
                        if length <= radius:
                            self.Q0[i][j] = power

    def power_reset(self, power_type='Q'):
        """Power reset.

        Resets the power matrix with `power_type` `'Q'` or `'Q0'`, which
        corresponds to the power temperature dependent and temperature
        independent, respectively.

        """
        # check the validity of inputs
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.Q = []
        self.Q0 = []
